#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import TYPE_CHECKING, List
from functools import reduce
from itertools import product

import numpy as np

from MDANSE.Framework.Units import measure
from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.Plotters.Plotter import Plotter

if TYPE_CHECKING:
    from qtpy.QtWidgets import QTextBrowser
    from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext, SingleDataset


class DatasetFormatter:
    """Independent of the GUI component, DatasetFormatter
    takes a PlottingContext and converts each of its
    datasets into text.
    The text can then be displayed by the plotter,
    outside of this class.
    """

    def __init__(self):
        self._plotting_context = None
        self._new_header = "# MDANSE Data \n"
        self._new_text = []
        self._is_preview = True
        self._preview_lines = 10
        self._preview_columns = 10
        self._rounding_prec = 5
        self._comment = "#"
        self._separator = " "

    def take_new_input(self, pc: "PlottingContext"):
        """The plotting context is passed from the GUI and
        contains datasets that have been selected by the user.
        This method converts each one of them to text and
        makes the result available to any GUI widget that
        may be used to display it.

        Parameters
        ----------
        pc : PlottingContext
            Data model containing the datasets for plotting

        Returns
        -------
        List[str]
            A list of (multi-line) text strings, one per dataset
        """
        self._plotting_context = pc
        self._new_text = []
        if self._plotting_context is None:
            return ["No data selected"]
        for name, databundle in self._plotting_context.datasets().items():
            dataset, _, _, _, _, axis_label = databundle
            if dataset._n_dim == 1:
                header, data = self.process_1D_data(
                    dataset, separator=self._separator, is_preview=self._is_preview
                )
                self._new_text.append(
                    self.join_for_gui(header, data, separator=self._separator)
                )
            elif dataset._n_dim == 2:
                header, data = self.process_2D_data(
                    dataset,
                    name,
                    separator=self._separator,
                    is_preview=self._is_preview,
                    main_axis=axis_label,
                )
                self._new_text.append(
                    self.join_for_gui(header, data, separator=self._separator)
                )
            else:
                header, data = self.process_ND_data(
                    dataset,
                    name,
                    separator=self._separator,
                    is_preview=self._is_preview,
                )
                self._new_text.append(
                    self.join_for_gui(header, data, separator=self._separator)
                )
        return self._new_text

    def datasets_for_csv(self):
        if self._plotting_context is None:
            return ["No data selected"]
        for name, databundle in self._plotting_context.datasets().items():
            dataset, _, _, _, _, _ = databundle
            if dataset._n_dim == 1:
                header, data = self.process_1D_data(dataset, separator=self._separator)
            elif dataset._n_dim == 2:
                header, data = self.process_2D_data(
                    dataset, name, separator=self._separator
                )
            else:
                header, data = self.process_ND_data(
                    dataset, name, separator=self._separator
                )
            yield header, data

    def make_dataset_header(self, dataset: "SingleDataset", comment_character="#"):
        """Extracts information related to the input dataset, and converts them
        into text. Returns a list of strings.

        Parameters
        ----------
        dataset : SingleDataset
            A dataset from MDANSE analysis results.
        comment_character : str, optional
            character used as prefix of the text header, by default '#'

        Returns
        -------
        List[str]
            list of header lines
        """
        lines = []
        lines.append(f"{comment_character} Dataset name: {dataset._name}")
        lines.append(f"{comment_character} from file {dataset._filename}")
        lines.append(
            f"{comment_character} Contains axes: "
            + ", ".join([str(axis) for axis in dataset.available_x_axes()])
        )
        lines.append(f"{comment_character} data unit is {dataset._data_unit}")
        return lines, comment_character

    def join_for_gui(
        self, header_lines: List[str], data_array: np.ndarray, separator=" "
    ):
        text_data = "\n".join(
            [
                separator.join([str(round(x, self._rounding_prec)) for x in line])
                for line in data_array
            ]
        )
        if self._is_preview:
            text_data += "\n..."
        new_header = "\n".join(header_lines)
        return new_header + "\n" + text_data

    def process_1D_data(
        self, dataset: "SingleDataset", separator=" ", is_preview=False
    ):
        """Formats a 1D array as a 2-column table with a commented header.
        The first column is determined using the information stored
        in the PlottingContext.

        Parameters
        ----------
        dataset : SingleDataset
            A SingleDataset read from an .MDA file (HDF5)
        separator : str, optional
            character(s) separating numbers in the output table, by default " "

        Returns
        -------
        str
            a data table with 2 columns in text format
        """
        header_lines, _ = self.make_dataset_header(
            dataset, comment_character=self._comment
        )
        best_unit, best_axis = dataset.longest_axis()
        xaxis_unit = self._plotting_context.get_conversion_factor(best_unit)
        try:
            conversion_factor = measure(1.0, best_unit, equivalent=True).toval(
                xaxis_unit
            )
        except:
            return f"Could not convert {best_unit} to {xaxis_unit}."
        else:
            header_lines.append(f"{self._comment} units of x axis here: {xaxis_unit}")
            header_lines.append(
                f"{self._comment} col1:{best_axis}:{xaxis_unit} col2:data:{dataset._data_unit}"
            )
            if is_preview:
                temp = np.vstack(
                    [
                        dataset._axes[best_axis][: self._preview_lines]
                        * conversion_factor,
                        dataset._data[: self._preview_lines],
                    ]
                ).T
            else:
                temp = np.vstack(
                    [dataset._axes[best_axis] * conversion_factor, dataset._data]
                ).T
            return header_lines, temp

    def process_2D_data(
        self,
        dataset: "SingleDataset",
        name: str,
        separator=" ",
        is_preview=False,
        main_axis=None,
    ):
        header_lines, comment_char = self.make_dataset_header(
            dataset, comment_character=self._comment
        )
        new_axes = {}
        new_axes_units = {}
        axis_numbers = {}
        flip_array = False
        for n, ax_key in enumerate(dataset.available_x_axes()):
            if main_axis is not None:
                if n > 0 and ax_key == main_axis:
                    flip_array = True
        for n, ax_key in enumerate(dataset.available_x_axes()):
            axis = dataset._axes[ax_key]
            axis_unit = dataset._axes_units[ax_key]
            new_unit = self._plotting_context.get_conversion_factor(axis_unit)
            conv_factor = measure(1.0, axis_unit, equivalent=True).toval(new_unit)
            new_axes[ax_key] = conv_factor * axis
            new_axes_units[ax_key] = new_unit
            axis_numbers[n] = ax_key
            LOG.debug(f"process_2D_data: axis {ax_key} has length {len(axis)}")
            if not flip_array:
                if n == 0:
                    header_lines.append(
                        f"{comment_char} first column is {ax_key} in units {new_unit}"
                    )
                else:
                    header_lines.append(
                        f"{comment_char} first row is {ax_key} in units {new_unit}"
                    )
            else:
                if n == 0:
                    header_lines.append(
                        f"{comment_char} first row is {ax_key} in units {new_unit}"
                    )
                else:
                    header_lines.append(
                        f"{comment_char} first column is {ax_key} in units {new_unit}"
                    )
        LOG.debug(f"Data shape: {dataset._data.shape}")
        if is_preview:
            nlines = self._preview_lines
            ncols = self._preview_columns
            nlines = min(nlines, len(new_axes[axis_numbers[0]]))
            ncols = min(ncols, len(new_axes[axis_numbers[1]]))
            temp = np.hstack(
                [
                    new_axes[axis_numbers[0]][:nlines].reshape((nlines, 1)),
                    dataset._data[:nlines, :ncols],
                ]
            )
            temp = np.vstack(
                [
                    np.concatenate([[0], new_axes[axis_numbers[1]][:ncols]]).reshape(
                        (1, ncols + 1)
                    ),
                    temp,
                ]
            )
        else:
            temp = np.hstack(
                [
                    new_axes[axis_numbers[0]].reshape((dataset._data.shape[0], 1)),
                    dataset._data,
                ]
            )
            temp = np.vstack(
                [
                    np.concatenate([[0], new_axes[axis_numbers[1]]]).reshape(
                        (1, dataset._data.shape[1] + 1)
                    ),
                    temp,
                ]
            )
        if not flip_array:
            return header_lines, temp
        return header_lines, temp.T

    def process_ND_data(
        self, dataset: "SingleDataset", name: str, separator=" ", is_preview=False
    ):
        header_lines, comment_char = self.make_dataset_header(
            dataset, comment_character=self._comment
        )
        new_axes = {}
        new_axes_units = {}
        axis_numbers = {}
        for n, ax_key in enumerate(dataset.available_x_axes()):
            axis = dataset._axes[ax_key]
            axis_unit = dataset._axes_units[ax_key]
            new_unit = self._plotting_context.get_conversion_factor(axis_unit)
            conv_factor = measure(1.0, axis_unit, equivalent=True).toval(new_unit)
            new_axes[ax_key] = conv_factor * axis
            new_axes_units[ax_key] = new_unit
            axis_numbers[n] = ax_key
            LOG.debug(f"process_ND_data: axis {ax_key} has length {len(axis)}")
        for ax_num, ax_key in enumerate(new_axes.keys()):
            header_lines.append(
                f"{comment_char} {ax_num} column is {ax_key} in units {new_axes_units[ax_key]}"
            )
        LOG.debug(f"Data shape: {dataset._data.shape}")
        temp = []
        ncols = len(new_axes) + 1
        ax_lengths = [len(new_axes[ax_num]) for ax_num in axis_numbers.values()]
        total_lines = reduce(lambda x, y: x * y, ax_lengths)
        if is_preview:
            counter = 0
            nlines = self._preview_lines
        else:
            counter = 0
            nlines = total_lines
        all_indices = product(*[range(ax_lengths[n]) for n in axis_numbers.keys()])
        while counter < nlines:
            array_index = all_indices.__next__()
            xvals = [
                new_axes[axis_numbers[axis_number]][index]
                for axis_number, index in enumerate(array_index)
            ]
            yval = dataset._data[array_index]
            temp.append(xvals + [yval])
            counter += 1
        return header_lines, np.vstack(temp)


class Text(Plotter):
    """The Text plotter is technically not a plotter, since
    it outputs text. The main advantage of imitating a plotter
    is that Text can use the GUI settings of units.
    In the end, PlottingContext matches the dataset to
    the x (y and z, optionally) axes in the data file,
    and the plotting interface supplies the user selection
    of physical units.
    """

    def __init__(self) -> None:
        super().__init__()
        self._figure = None
        self._current_colours = []
        self._active_curves = []
        self._backup_curves = []
        self._backup_limits = []
        self._formatter = DatasetFormatter()
        self._curve_limit_per_dataset = 12
        self._pc_backup = None
        self.height_max, self.length_max = 0.0, 0.0

    def clear(self, figure: "QTextBrowser" = None):
        """Optionally sets the input to be the new
        text output widget, and then clears the
        currently stored text widget instance.

        Parameters
        ----------
        figure : QTextBrowser, optional
            optionally, new QTextBrowser to be used
        """
        LOG.debug("Text.clear stared")
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            return
        target.clear()
        LOG.debug("Text.clear finished")

    def adjust_formatter(
        self,
        preview=False,
        preview_lines=10,
        preview_columns=10,
        rounding=15,
        separator=" ",
        comment="#",
    ):
        if self._formatter is None or self._pc_backup is None or self._figure is None:
            return
        self._formatter._is_preview = preview
        self._formatter._preview_lines = preview_lines
        self._formatter._preview_columns = preview_columns
        self._formatter._rounding_prec = rounding
        self._formatter._separator = separator
        self._formatter._comment = comment
        self.plot(self._pc_backup, self._figure)

    def get_figure(self, figure: "QTextBrowser" = None):
        """Used for both updating and getting the output widget
        which will be used for displaying the output text.

        Parameters
        ----------
        figure : QTextBrowser, optional
            New widget to be used for displaying text.
            If None, the currently stored widget will be used.

        Returns
        -------
        QTextBrowser or None
            If no widget reference can be used, returns None.
            Otherwise, returns a widget instance that should be used
            for displaying text.
        """
        LOG.debug("Text.get_figure stared")
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            LOG.error(f"PlottingContext can't plot to {target}")
            return
        target.clear()
        LOG.debug("Text.get_figure finished")
        return target

    def apply_settings(self, plotting_context: "PlottingContext", colours=None):
        """Not relevant to the Text plotter, added for compatibility

        Parameters
        ----------
        plotting_context : PlottingContext
            ignored
        colours : _type_, optional
            ignored
        """
        LOG.debug("Text.apply_settings called. Doing nothing")

    def plot(
        self,
        plotting_context: "PlottingContext",
        figure: "QTextBrowser" = None,
        colours=None,
        update_only=False,
        toolbar=None,
    ):
        """In the Text plotter, the plot method displays the data as text.
        Many arguments are included just for compatibility with the plotter.

        Parameters
        ----------
        plotting_context : PlottingContext
            Data model containing the data sets to be shown
        figure : QTextBrowser, optional
            Target widget, an instance of QTextBrowser
        colours : _type_, optional
            ignored here
        update_only : bool, optional
            ignored
        toolbar : _type_, optional
            ignored
        """
        LOG.debug("Text.plot stared")
        target = self.get_figure(figure)
        if target is None:
            return
        if toolbar is not None:
            self._toolbar = toolbar
        self._pc_backup = plotting_context
        self._figure = target
        xaxis_unit = None
        self._active_curves = []
        self._backup_curves = []
        self.apply_settings(plotting_context, colours)
        self.height_max, self.length_max = 0.0, 0.0
        if plotting_context.set_axes() is None:
            LOG.debug("Axis check failed.")
            return
        if len(plotting_context.datasets()) == 0:
            target.clear()
            return
        self._formatter.take_new_input(plotting_context)
        target.setText("\n".join(self._formatter._new_text))



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from qtpy.QtWidgets import (
    QDoubleSpinBox,
    QComboBox,
    QItemDelegate,
    QStyledItemDelegate,
    QColorDialog,
    QApplication,
    QStyle,
    QStyleOptionProgressBar,
)
from qtpy.QtCore import Signal, Slot, Qt
from qtpy.QtGui import QColor


class ColourPicker(QStyledItemDelegate):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def createEditor(self, parent, option, index):
        dialog = QColorDialog(parent)
        return dialog

    def setEditorData(self, editor, index):
        colour_string = index.data()
        color = QColor(colour_string)
        editor.setCurrentColor(color)

    def setModelData(self, editor, model, index):
        if editor.result() == QColorDialog.DialogCode.Accepted:
            color = editor.currentColor()
            colour_string = color.toRgb()
            model.setData(index, colour_string)
            model.setData(index, color, role=Qt.ItemDataRole.BackgroundRole)


class MainAxisCombo(QItemDelegate):

    def __init__(self, *args, **kwargs) -> None:
        self._items = []
        super().__init__(*args, **kwargs)

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        text = index.model().data(index, Qt.DisplayRole)
        try:
            i = self._items.index(text)
        except ValueError:
            i = 0
        editor.setCurrentIndex(i)
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.currentText())

    def createEditor(self, parent, option, index):
        combo = QComboBox(parent)
        model = index.model()
        key = index.model().data(index, Qt.ItemDataRole.UserRole)
        dataset = model._datasets[key]
        self._items = [str(x) for x in dataset._axes.keys()]
        combo.addItems(self._items)
        combo.currentIndexChanged.connect(self.currentIndexChanged)
        return combo

    @Slot()
    def currentIndexChanged(self):
        self.commitData.emit(self.sender())


class MplStyleCombo(QItemDelegate):

    def __init__(self, *args, mpl_items=None, **kwargs) -> None:
        self._items = list([str(x) for x in mpl_items.keys()])
        super().__init__(*args, **kwargs)

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        text = index.model().data(index, Qt.DisplayRole)
        try:
            i = self._items.index(text)
        except ValueError:
            i = 0
        editor.setCurrentIndex(i)
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.currentText())

    def createEditor(self, parent, option, index):
        combo = QComboBox(parent)
        li = []
        for item in self._items:
            li.append(item)
        combo.addItems(li)
        combo.currentIndexChanged.connect(self.currentIndexChanged)
        return combo

    @Slot()
    def currentIndexChanged(self):
        self.commitData.emit(self.sender())


class RadiusSpinBox(QItemDelegate):

    def __init__(self, *args, mpl_items=None, **kwargs) -> None:
        self._minimum = 0.0
        self._step = 0.01
        super().__init__(*args, **kwargs)

    def setEditorData(self, editor, index):
        current_value = float(index.data())
        editor.setValue(current_value)

    def setModelData(self, editor, model, index):
        model.setData(index, str(round(editor.value(), 2)))

    def createEditor(self, parent, option, index):
        sbox = QDoubleSpinBox(parent)
        sbox.setMinimum(self._minimum)
        sbox.setSingleStep(self._step)
        sbox.valueChanged.connect(self.valueChanged)
        return sbox

    @Slot()
    def valueChanged(self):
        self.commitData.emit(self.sender())


class ProgressDelegate(QItemDelegate):

    progress_role = Qt.UserRole + 1000

    def paint(self, painter, option, index):
        progress = index.data(self.progress_role)
        progress_max = max(index.data(self.progress_role + 1) - 1, 1)
        try:
            int(progress)
        except:
            progress = 0
        opt = QStyleOptionProgressBar()
        opt.rect = option.rect
        opt.minimum = 0
        opt.maximum = progress_max
        opt.progress = progress
        percentage = min(round(100 * progress / progress_max, 2), 100.0)
        opt.text = "{}%".format(percentage)
        opt.textVisible = True
        QApplication.style().drawControl(QStyle.CE_ProgressBar, opt, painter)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QLineEdit, QDoubleSpinBox
from qtpy.QtGui import QDoubleValidator

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class FloatWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            default_option = float(self._configurator.default)
        except ValueError:
            default_option = 0.0
        if self._configurator.choices:
            field = QDoubleSpinBox(self._base)
            field.setMinimum(self._configurator.choices[0])
            field.setMaximum(self._configurator.choices[-1])
            if len(self._configurator.choices) > 1:
                field.setSingleStep(
                    self._configurator.choices[1] - self._configurator.choices[0]
                )
            field.setValue(default_option)
        else:
            field = QLineEdit(self._base)
            validator = QDoubleValidator(field)
            minval, maxval = self._configurator.mini, self._configurator.maxi
            if minval is not None:
                validator.setBottom(minval)
            if maxval is not None:
                validator.setTop(maxval)
            field.setValidator(validator)
            field.setText(str(default_option))
            field.textChanged.connect(self.updateValue)
            field.setPlaceholderText(str(default_option))
        field.setToolTip(self._tooltip)
        self._layout.addWidget(field)
        self._field = field
        self._default_value = default_option
        self.default_labels()
        self.update_labels()
        self.updateValue()
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "A single floating-point number"
        field.setToolTip(tooltip_text)

    def default_labels(self):
        """Each Widget should have a default tooltip and label,
        which will be set in this method, unless specific
        values are provided in the settings of the job that
        is being configured."""
        if self._label_text == "":
            self._label_text = "FloatWidget"
        if self._tooltip == "":
            self._tooltip = "A single floating-point number"

    def get_widget_value(self):
        """Collect the results from the input widgets and return the value."""
        strval = self._field.text().strip()
        if len(strval) < 1:
            self._empty = True
            return self._default_value
        else:
            self._empty = False
        return strval



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import os
import tomlkit
import copy
from tomlkit.parser import ParseError
from tomlkit.toml_file import TOMLFile

from qtpy.QtCore import Slot, Signal, QModelIndex, Qt
from qtpy.QtWidgets import QListView, QAbstractItemView, QMenu
from qtpy.QtGui import QContextMenuEvent, QStandardItem

from MDANSE import PLATFORM
from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.Visualisers.InstrumentDetails import (
    InstrumentDetails,
    SimpleInstrument,
)


class InstrumentList(QListView):
    item_details = Signal(object)
    item_name = Signal(str)
    error = Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.clicked.connect(self.item_picked)
        self._current_instrument = None
        self._all_instruments = set()
        self._backup_instruments = {}

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        index = self.indexAt(event.pos())
        if index.row() == -1:
            # block right click when it's not on a trajectory
            return
        model = self.model()
        item = model.itemData(index)
        menu = QMenu()
        self.populateMenu(menu, item)
        menu.exec_(event.globalPos())

    def populateMenu(self, menu: QMenu, item: QStandardItem):
        has_backup = False
        instrument = self.model()._nodes[item[257]]
        if instrument is None:
            return
        if instrument._name in self._backup_instruments.keys():
            has_backup = True
        for action, method in [
            ("Delete instrument", self.deleteNode),
        ]:
            temp_action = menu.addAction(action)
            temp_action.triggered.connect(method)
            if has_backup:
                temp_action.setEnabled(False)
        for action, method in [
            ("Duplicate instrument", self.duplicateNode),
        ]:
            temp_action = menu.addAction(action)
            temp_action.triggered.connect(method)
        for action, method in [
            ("Restore original parameters", self.restoreNode),
        ]:
            temp_action = menu.addAction(action)
            temp_action.triggered.connect(method)
            if not has_backup:
                temp_action.setEnabled(False)

    @Slot()
    def deleteNode(self):
        model = self.model()
        index = self.currentIndex()
        model.removeRow(index.row())
        self.item_details.emit(None)

    @Slot()
    def duplicateNode(self):
        model = self.model()
        index = self.currentIndex()
        node_number = model.itemFromIndex(index).data()
        instrument = model._nodes[node_number]
        new_instrument_name = instrument._name + " (Copy)"
        new_instrument = self.add_instrument(new_instrument_name)
        for line in new_instrument.inputs():
            attr_name = line[0]
            setattr(new_instrument, attr_name, getattr(instrument, attr_name, None))
        new_instrument._name = instrument._name + " (Copy)"
        if new_instrument._list_item is not None:
            new_instrument._list_item.setText(new_instrument_name)

    @Slot()
    def restoreNode(self):
        model = self.model()
        index = self.currentIndex()
        node_number = model.itemFromIndex(index).data()
        instrument = model._nodes[node_number]
        name = instrument._name
        # TBC
        try:
            backup = self._backup_instruments[name]
        except KeyError:
            LOG.error(f"Backup instrument not found: {name}")
        else:
            for line in instrument.inputs():
                attr_name = line[0]
                setattr(instrument, attr_name, getattr(backup, attr_name, None))
            self.item_details.emit(instrument)

    @Slot(QModelIndex)
    def item_picked(self, index: QModelIndex):
        model = self.model()
        node_number = model.itemFromIndex(index).data()
        instrument = model._nodes[node_number]
        self._current_instrument = instrument
        self.item_details.emit(instrument)

    @Slot()
    def resend_item(self):
        self.item_details.emit(self._current_instrument)

    def add_instrument(self, optional_name=None) -> SimpleInstrument:
        model = self.model()
        if model is None:
            return
        new_instrument = SimpleInstrument()
        LOG.debug(f"New instrument, name: {new_instrument._name}")
        if optional_name is None:
            new_name = new_instrument._name
        else:
            new_name = optional_name
        model.append_object_and_embed((new_instrument, new_name))
        self._all_instruments.add(new_name)
        new_instrument.update_item()
        return new_instrument

    def connect_to_visualiser(self, visualiser: InstrumentDetails) -> None:
        """Connect to a visualiser.

        Parameters
        ----------
        visualiser : View3D or TrajectoryInfo
            A visualiser to connect to this view.
        """
        try:
            self.item_details.connect(visualiser.update_panel)
        except Exception as e:
            LOG.error(f"Failed to connect InstrumentList to visualiser: {e}")

    def introduce_empty_instrument(self):
        self.model().append_object((None, "No instrument (default settings)"))

    def load_from_file(self, filename: str, keep_backups=False):
        source_file = TOMLFile(filename)
        try:
            tomldoc = source_file.read()
        except ParseError:
            LOG.error(f"File {filename} could not be parsed - TOML error")
            return
        for key in tomldoc.keys():
            if key in self._all_instruments:
                LOG.warning(
                    f"{key} already on instrument list. Overwriting from {filename}"
                )
                for node in self.model()._nodes.values():
                    new_instrument = node
                    if new_instrument is None:
                        continue
                    if new_instrument._name == key:
                        break
            else:
                new_instrument = self.add_instrument(str(key))
            if new_instrument is None:
                return
            if keep_backups:
                backup_instrument = SimpleInstrument()
            instrument_params = tomldoc[key]
            for input in SimpleInstrument.inputs():
                param_key = input[0]
                try:
                    new_value = instrument_params[param_key]
                except KeyError:
                    continue
                else:
                    setattr(new_instrument, param_key, new_value)
                    if keep_backups:
                        setattr(backup_instrument, param_key, new_value)
            if keep_backups:
                self._backup_instruments[str(key)] = backup_instrument
                new_instrument._name_is_fixed = True

    @Slot()
    def save_to_file(self, filename: str = None):
        if filename is None:
            filename = os.path.join(
                PLATFORM.application_directory(), "InstrumentDefinitions.toml"
            )
        LOG.debug(f"save to file: {filename}")
        target_file = TOMLFile(filename)
        newdoc = tomlkit.document()
        model = self.model()
        for instrument in model._nodes.values():
            if instrument is None:
                continue
            table = tomlkit.table()
            for input in SimpleInstrument.inputs():
                param_key = input[0]
                table[param_key] = getattr(instrument, param_key, "")
            newdoc[instrument._name] = table
        target_file.write(newdoc)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import numpy as np
from qtpy.QtGui import QStandardItemModel, QStandardItem
from qtpy.QtWidgets import QTableView, QColorDialog

from MDANSE_GUI.MolecularViewer.readers.i_reader import IReader


class TrajectoryAtomData(QStandardItemModel):
    """This object will store the information about the atoms
    present in the trajectory, allowing the user to customise
    the way the trajectory is visualised.
    It is a subclass of QStandardItemModel, allowing the information
    to be used with a QTableView widget.
    """

    def __init__(self, *args, **kwargs):
        super(TrajectoryAtomData, self).__init__(*args, **kwargs)
        self._viewer = None
        self._reader = None

    def setViewer(self, viewer):
        """Associates the data model with an instance of the
        viewer widget, so new settings can be passed to the
        visualiser.

        Arguments:
            viewer -- an instance of the MolecularViewer widget
        """
        self._viewer = viewer
        self.itemChanged.connect(self._viewer._new_atom_parameters)

    def setReader(self, ireader: IReader):
        """Assigns an instance of IReader to the data model,
        so that information about the atoms can be extracted
        from the reader.

        Arguments:
            ireader -- a trajectory reader for the MolecularViewer.
        """
        self._reader = ireader

    def parseInformation(self, unique=True):
        if self._reader is None or self._viewer is None:
            return
        atom_names = self._reader.atom_names
        atom_types = self._reader.atom_types
        atom_ids = self._reader.atom_ids
        grouped_by_type = {}
        name_by_type = {}
        colour_by_type = {}
        radius_by_type = {}
        unique_types = np.unique(atom_types)
        for type in unique_types:
            crit = np.where(atom_types == type)
            grouped_by_type[type] = atom_ids[crit]
            name_by_type[type] = atom_names[crit][0]



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import glob
import importlib
import os

current_path, _ = os.path.split(__file__)

modnames = []
fnames = glob.glob(current_path + "/*.py")
for fname in fnames:
    _, newname = os.path.split(fname)
    newname = newname.split(".py")[0]
    modnames.append(newname)
globdict = globals()


for name in modnames:
    if name in ["__init__"]:
        continue
    try:
        tempmod = importlib.import_module("." + name, "MDANSE_GUI.InputWidgets")
    except ModuleNotFoundError:
        continue
    tempobject = getattr(tempmod, name)
    globdict[name] = tempobject
    del tempmod  # optionally delete the reference to the parent module



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#


from qtpy.QtWidgets import (
    QLabel,
    QLineEdit,
    QPushButton,
    QComboBox,
)
from qtpy.QtCore import Slot
from qtpy.QtGui import QDoubleValidator

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase
from MDANSE_GUI.Widgets.ResolutionDialog import ResolutionDialog
from MDANSE_GUI.Widgets.ResolutionWidget import widget_text_map


reverse_text_map = {value: key for key, value in widget_text_map.items()}


init_parameters = {
    "ideal": {},
    "Gaussian": {"mu": 0.0, "sigma": 1.0},
    "Lorentzian": {"mu": 0.0, "sigma": 1.0},
    "triangular": {"mu": 0.0, "sigma": 1.0},
    "square": {"mu": 0.0, "sigma": 1.0},
    "pseudo-Voigt": {
        "mu_gaussian": 0.0,
        "sigma_gaussian": 1.0,
        "eta": 0.5,
        "mu_lorentzian": 0.0,
        "sigma_lorentzian": 1.0,
    },
}


class InstrumentResolutionWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        kwargs["layout_type"] = "QGridLayout"
        super().__init__(*args, **kwargs)
        # self._layout.addWidget(QLabel("Resolution function", self._base), 0, 0)
        self._type_combo = QComboBox(parent=self._base)
        self._dialog_button = QPushButton("Helper Dialog", self._base)
        self._dialog_button.clicked.connect(self.helper_dialog)
        self._dialog_button.setEnabled(True)
        self.helper = ResolutionDialog(self._base)
        self.helper._panel.parameters_changed.connect(self.set_parameters_from_dialog)
        self._layout.addWidget(self._type_combo, 0, 1)
        self._layout.addWidget(self._dialog_button, 0, 0)
        self._labels = []
        self._fields = []
        self._defaults = []
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "The peak function used for smearing/smooting the results. Pick 'ideal' for no smearing."
        # first row
        for num in range(1, 3):
            label = QLabel("", parent=self._base)
            field = QLineEdit("N/A", parent=self._base)
            field.setEnabled(False)
            self._layout.addWidget(label, 0, 2 * num)
            self._layout.addWidget(field, 0, 2 * num + 1)
            self._labels.append(label)
            self._fields.append(field)
            val = QDoubleValidator(field)
            field.setValidator(val)
            self._defaults.append(0.0)
        # first row
        for num in range(0, 3):
            label = QLabel("", parent=self._base)
            field = QLineEdit("N/A", parent=self._base)
            field.setEnabled(False)
            self._layout.addWidget(label, 1, 2 * num)
            self._layout.addWidget(field, 1, 2 * num + 1)
            self._labels.append(label)
            self._fields.append(field)
            val = QDoubleValidator(field)
            if num == 0:  # this is the pseudoVoigt eta
                val.setBottom(0.0)
                val.setTop(1.0)
            field.setValidator(val)
            self._defaults.append(0.0)
        self._type_combo.addItems([str(x) for x in init_parameters.keys()])
        self._type_combo.setEditable(False)
        self._type_combo.currentTextChanged.connect(self.change_function)
        for field in self._fields:
            field.textChanged.connect(self.updateValue)
            field.setToolTip(tooltip_text)
        self.updateValue()

    def configure_using_default(self):
        """This is too complex to have a default value"""

    def set_field_values(self, new_params: dict):
        np = list(new_params.items())
        for index in range(5):
            try:
                kv = np[index]
                self._fields[index].setEnabled(True)
                self._labels[index].setText(str(kv[0]))
                self._fields[index].setText(str(kv[1]))
                self._fields[index].setPlaceholderText(str(kv[1]))
                self._defaults[index] = float(kv[1])
            except IndexError:
                self._labels[index].setText("")
                self._fields[index].setText("")
                self._fields[index].setPlaceholderText("N/A")
                self._fields[index].setEnabled(False)
                self._defaults[index] = 0.0

    @Slot(str)
    def change_function(self, function: str, optional_parameters: dict = None):
        if optional_parameters is None:
            if function in init_parameters.keys():
                new_params = init_parameters[function]
            else:
                new_params = init_parameters[widget_text_map[function]]
        else:
            new_params = optional_parameters
        self._type_combo.blockSignals(True)
        if function in widget_text_map.keys():
            self._type_combo.setCurrentText(widget_text_map[function])
        elif function in reverse_text_map.keys():
            self._type_combo.setCurrentText(reverse_text_map[function])
        else:
            self._type_combo.setCurrentText(function)
        self._type_combo.blockSignals(False)
        self.set_field_values(new_params)

    def get_widget_value(self):
        function = widget_text_map[self._type_combo.currentText()]
        params = {}
        for index in range(5):
            key = self._labels[index].text()
            if len(key) > 0:
                try:
                    value = float(self._fields[index].text())
                except ValueError:
                    value = self._defaults[index]
                else:
                    self._fields[index].setStyleSheet("")
                params[key] = value
        self.helper._panel.update_fields((function, params))
        return (function, params)

    @Slot(dict)
    def set_parameters_from_dialog(self, input: dict):
        peak_function = input.get("function", None)
        if peak_function is None:
            return
        self._type_combo.setCurrentText(peak_function)
        temp_parameters = init_parameters[peak_function]
        for key in temp_parameters.keys():
            temp_parameters[key] = input.get(key, 0.0)
        self.set_field_values(temp_parameters)

    @Slot()
    def helper_dialog(self):
        if self.helper.isVisible():
            self.helper.close()
        else:
            self.helper.show()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import Optional

import numpy as np
from qtpy.QtWidgets import (
    QPushButton,
    QFileDialog,
    QLabel,
    QWidget,
    QLineEdit,
    QComboBox,
    QGridLayout,
    QTextEdit,
)
from qtpy.QtCore import Signal, Slot, Qt
from qtpy.QtGui import QStandardItem, QDoubleValidator, QIntValidator

from MDANSE.MLogging import LOG

from MDANSE_GUI.Widgets.VectorWidget import VectorWidget
from MDANSE_GUI.Tabs.Visualisers.InstrumentInfo import InstrumentInfo
from MDANSE_GUI.Tabs.Visualisers.InstrumentInfo import SimpleInstrument


class InstrumentDetails(QWidget):

    instrument_details_changed = Signal(int)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setLayout(QGridLayout(self))
        self._widgets = {}
        self._values = {}
        self._current_instrument = None
        self.populate_panel(SimpleInstrument)
        self.add_visualiser()
        for attr in ["_sample", "_technique"]:
            self._widgets[attr].currentTextChanged.connect(self.reset_qvector_combobox)

    def add_visualiser(self):
        layout = self.layout()
        next_row = layout.rowCount()
        self._inner_visualiser = InstrumentInfo(self)
        layout.addWidget(self._inner_visualiser, next_row, 0, 1, 2)

    def create_widget(self, entry):
        layout = self.layout()
        next_row = layout.rowCount()
        attribute = entry[0]
        widget = entry[1]
        value = entry[2]
        label = attribute.strip(" _")
        if widget == "QComboBox":
            instance = QComboBox(self)
            instance.addItems(getattr(SimpleInstrument, value))
            instance.currentTextChanged.connect(self.update_values)
        elif widget == "QLineEdit":
            instance = QLineEdit(self)
            if value == "float":
                instance.setValidator(QDoubleValidator())
                instance.setText("1.0")
                instance.setPlaceholderText("1.0")
            elif value == "int":
                instance.setValidator(QIntValidator())
                instance.setText("1")
                instance.setPlaceholderText("1")
            else:
                instance.setText("text")
            instance.textChanged.connect(self.update_values)
        elif widget == "VectorWidget":
            instance = VectorWidget(self)
            instance.set_value([0, 0, 0])
            instance.value_changed.connect(self.update_values)
        qlabel = QLabel(label, self)
        layout.addWidget(qlabel, next_row, 0)
        layout.addWidget(instance, next_row, 1)
        self._widgets[attribute] = instance
        for widg in self._widgets.values():
            widg.setEnabled(False)

    @Slot()
    def reset_qvector_combobox(self):
        if self._current_instrument is None:
            return
        qvec_combo = self._widgets["_qvector_type"]
        last_option = qvec_combo.currentText()
        new_options = self._current_instrument.filter_qvector_generator()
        qvec_combo.clear()
        if len(new_options) == 0:
            return
        qvec_combo.addItems(new_options)
        if last_option in new_options:
            qvec_combo.setCurrentText(last_option)
        else:
            qvec_combo.setCurrentText(new_options[0])

    @Slot()
    def update_values(self):
        for key, value in self._widgets.items():
            try:
                new_val = value.text()
            except AttributeError:
                try:
                    new_val = value.currentText()
                except AttributeError:
                    new_val = None
            if len(new_val) == 0:
                try:
                    new_val = value.placeholderText()
                except AttributeError:
                    new_val = ""
            self._values[key] = new_val
        self.commit_changes()
        self.toggle_axis_fields()

    def commit_changes(self):
        if self._current_instrument is None:
            return
        for key, value in self._values.items():
            setattr(self._current_instrument, key, value)
        self._current_instrument._configured = True
        self._current_instrument.update_item()
        self._inner_visualiser.update_panel(self._current_instrument)
        self.instrument_details_changed.emit(self._current_instrument._model_index)

    def populate_panel(self, instrument: SimpleInstrument):
        for entry in instrument.inputs():
            self.create_widget(entry)

    def toggle_axis_fields(self):
        if self._current_instrument is None:
            return
        for name, widget in self._widgets.items():
            if "axis_1" in name:
                if not (
                    "Linear" in self._current_instrument._qvector_type
                    or "Circular" in self._current_instrument._qvector_type
                ):
                    widget.setEnabled(False)
                else:
                    widget.setEnabled(True)
            if "axis_2" in name:
                if not "Circular" in self._current_instrument._qvector_type:
                    widget.setEnabled(False)
                else:
                    widget.setEnabled(True)

    def clear_panel(self):
        for widget in self._widgets.values():
            widget.setEnabled(False)

    @Slot(object)
    def update_panel(self, instrument_instance):
        if instrument_instance is None:
            self._current_instrument = None
            self.clear_panel()
            return
        for widg in self._widgets.values():
            widg.setEnabled(True)
        self._values = {}
        self._current_instrument = None
        for key, widget in self._widgets.items():
            new_value = getattr(instrument_instance, key, "Nothing!")
            try:
                widget.setText(str(new_value))
            except AttributeError:
                try:
                    widget.setCurrentText(str(new_value))
                except Exception as e:
                    LOG.error(f"In InstrumentDetails: {e}")
        self._current_instrument = instrument_instance
        self.update_values()
        self.reset_qvector_combobox()
        if instrument_instance._name_is_fixed:
            self._widgets["_name"].setEnabled(False)
        else:
            self._widgets["_name"].setEnabled(True)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import logging
import sys
import os
import time

from qtpy.QtWidgets import QApplication, QSplashScreen, QStyleFactory
from qtpy.QtCore import QSettings, Qt, QTimer
from qtpy.QtGui import QPixmap, QIcon

from MDANSE.MLogging import LOG, FMT

from MDANSE_GUI.TabbedWindow import TabbedWindow


os.environ["OMP_NUM_THREADS"] = "1"


# an additonal section which will pass exception information to the logger
def catch_exceptions(t, val, tb):
    LOG.error(f"EXCEPTION:\n{t}\n{val}\n{tb}")
    old_hook(t, val, tb)


old_hook = sys.excepthook
sys.excepthook = catch_exceptions
# end of exception handling part.


def startGUI(some_args):
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel("INFO")
    stream_handler.setFormatter(FMT)
    LOG.addHandler(stream_handler)

    app = QApplication(some_args)
    app.setStyle(QStyleFactory.create("Fusion"))

    path = os.path.dirname(os.path.abspath(__file__))
    app.setWindowIcon(QIcon(os.path.join(path, "Icons/MDANSE.ico")))

    settings = QSettings(
        "ISIS Neutron and Muon Source", "MDANSE for Python 3", parent=app
    )

    path = os.path.dirname(os.path.abspath(__file__))
    splash_img = QPixmap(os.path.join(path, "Resources/splash.png"))
    splash = QSplashScreen(splash_img, Qt.WindowStaysOnTopHint)
    splash.show()
    t0 = time.time()
    root = TabbedWindow(
        parent=None, title="MDANSE for Python 3", settings=settings, app_instance=app
    )
    root.show()
    t1 = time.time()
    # only allow the splash screen to stay up for maximum of 2s if the
    # GUI loads fast but closes immediately if GUI was slow to load
    QTimer.singleShot(
        max([0, round(2000 - 1000 * (t1 - t0))]), lambda: splash.finish(root)
    )

    app.exec()  # once this is done, the GUI has its event loop running.
    # no more Python scripting now, we are in the event loop.


if __name__ == "__main__":
    startGUI(sys.argv)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import (
    QLineEdit,
    QPushButton,
    QDialog,
    QCheckBox,
    QVBoxLayout,
    QHBoxLayout,
    QGroupBox,
    QLabel,
    QTextEdit,
    QWidget,
)
from MDANSE.Framework.AtomSelector import Selector
from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase
from MDANSE_GUI.Tabs.Visualisers.View3D import View3D
from MDANSE_GUI.MolecularViewer.MolecularViewer import MolecularViewerWithPicking
from MDANSE.Framework.InputData.HDFTrajectoryInputData import HDFTrajectoryInputData
from .CheckableComboBox import CheckableComboBox


class SelectionHelper(QDialog):
    """Generates a string that specifies the atom selection.

    Attributes
    ----------
    _helper_title : str
        The title of the helper dialog window.
    _cbox_text : dict
        The dictionary that maps the selector settings to text used in
        the helper dialog.
    """

    _helper_title = "Atom selection helper"
    _cbox_text = {
        "all": "All atoms (excl. dummy atoms):",
        "dummy": "All dummy atoms:",
        "hs_on_heteroatom": "Hs on heteroatoms:",
        "primary_amine": "Primary amine groups:",
        "hydroxy": "Hydroxy groups:",
        "methyl": "Methyl groups:",
        "phosphate": "Phosphate groups:",
        "sulphate": "Sulphate groups:",
        "thiol": "Thiol groups:",
        "water": "Water molecules:",
        "hs_on_element": "Hs on elements:",
        "element": "Elements:",
        "name": "Atom name:",
        "fullname": "Atom fullname:",
        "index": "Indexes:",
    }

    def __init__(
        self,
        selector: Selector,
        traj_data: tuple[str, HDFTrajectoryInputData],
        field: QLineEdit,
        parent,
        *args,
        **kwargs,
    ):
        """
        Parameters
        ----------
        selector : Selector
            The MDANSE selector initialized with the current chemical
            system.
        traj_data : tuple[str, HDFTrajectoryInputData]
            A tuple of the trajectory data used to load the 3D viewer.
        field : QLineEdit
            The QLineEdit field that will need to be updated when
            applying the setting.
        """
        super().__init__(parent, *args, **kwargs)
        self.setWindowTitle(self._helper_title)

        self.selector = selector
        self._field = field
        self.full_settings = self.selector.full_settings

        self.selection_textbox = QTextEdit()
        self.selection_textbox.setReadOnly(True)

        mol_view = MolecularViewerWithPicking()
        mol_view.picked_atoms_changed.connect(self.update_from_3d_view)
        self.view_3d = View3D(mol_view)
        self.view_3d.update_panel(traj_data)

        layouts = self.create_layouts()

        bottom = QHBoxLayout()
        for button in self.create_buttons():
            bottom.addWidget(button)

        layouts[-1].addLayout(bottom)

        helper_layout = QHBoxLayout()
        for layout in layouts:
            helper_layout.addLayout(layout)

        self.setLayout(helper_layout)
        self.update_others()

        self.all_selection = True

    def closeEvent(self, a0):
        """Hide the window instead of closing. Some issues occur in the
        3D viewer when it is closed and then reopened.
        """
        a0.ignore()
        self.hide()

    def create_buttons(self) -> list[QPushButton]:
        """
        Returns
        -------
        list[QPushButton]
            List of push buttons to add to the last layout from
            create_layouts.
        """
        apply = QPushButton("Use Setting")
        reset = QPushButton("Reset")
        close = QPushButton("Close")
        apply.clicked.connect(self.apply)
        reset.clicked.connect(self.reset)
        close.clicked.connect(self.close)
        return [apply, reset, close]

    def create_layouts(self) -> list[QVBoxLayout]:
        """
        Returns
        -------
        list[QVBoxLayout]
            List of QVBoxLayout to add to the helper layout.
        """
        layout_3d = QVBoxLayout()
        layout_3d.addWidget(self.view_3d)

        left = QVBoxLayout()
        for widget in self.left_widgets():
            left.addWidget(widget)

        right = QVBoxLayout()
        for widget in self.right_widgets():
            right.addWidget(widget)

        return [layout_3d, left, right]

    def right_widgets(self) -> list[QWidget]:
        """
        Returns
        -------
        list[QWidget]
            List of QWidgets to add to the right layout from
            create_layouts.
        """
        return [self.selection_textbox]

    def left_widgets(self) -> list[QWidget]:
        """
        Returns
        -------
        list[QWidget]
            List of QWidgets to add to the left layout from
            create_layouts.
        """
        match_exists = self.selector.match_exists

        select = QGroupBox("selection")
        select_layout = QVBoxLayout()

        self.check_boxes = []
        self.combo_boxes = []

        for k, v in self.full_settings.items():

            if isinstance(v, bool):
                check_layout = QHBoxLayout()
                checkbox = QCheckBox()
                checkbox.setChecked(v)
                checkbox.setLayoutDirection(Qt.RightToLeft)
                label = QLabel(self._cbox_text[k])
                checkbox.setObjectName(k)
                checkbox.stateChanged.connect(self.update_others)
                if not match_exists[k]:
                    checkbox.setEnabled(False)
                    label.setStyleSheet("color: grey;")
                self.check_boxes.append(checkbox)
                check_layout.addWidget(label)
                check_layout.addWidget(checkbox)
                select_layout.addLayout(check_layout)

            elif isinstance(v, dict):
                combo_layout = QHBoxLayout()
                combo = CheckableComboBox()
                items = [str(i) for i in v.keys() if match_exists[k][i]]
                # we blocksignals here as there can be some
                # performance issues with a large number of items
                combo.model().blockSignals(True)
                combo.addItems(items)
                combo.model().blockSignals(False)
                combo.setObjectName(k)
                combo.model().dataChanged.connect(self.update_others)
                label = QLabel(self._cbox_text[k])
                if len(items) == 0:
                    combo.setEnabled(False)
                    label.setStyleSheet("color: grey;")
                self.combo_boxes.append(combo)
                combo_layout.addWidget(label)
                combo_layout.addWidget(combo)
                select_layout.addLayout(combo_layout)

        select.setLayout(select_layout)
        return [select]

    def update_others(self) -> None:
        """Using the checkbox and combobox widgets: update the settings,
        get the selection and update the textedit box with details of
        the current selection and the 3d view to match the selection.
        """
        for check_box in self.check_boxes:
            self.full_settings[check_box.objectName()] = check_box.isChecked()
        for combo_box in self.combo_boxes:
            for item in combo_box.getItems():
                txt = item.text()
                if combo_box.objectName() == "index":
                    key = int(txt)
                else:
                    key = txt
                self.full_settings[combo_box.objectName()][key] = (
                    item.checkState() == Qt.Checked
                )

        self.selector.update_settings(self.full_settings)
        idxs = self.selector.get_idxs()
        self.view_3d._viewer.change_picked(idxs)
        self.update_selection_textbox(idxs)

    def update_from_3d_view(self, selection: set[int]) -> None:
        """A selection/deselection was made in the 3d view, update the
        check_boxes, combo_boxes and textbox.

        Parameters
        ----------
        selection : set[int]
            Selection indexes from the 3d view.
        """
        self.selector.update_with_idxs(selection)
        self.full_settings = self.selector.full_settings
        self.update_selection_widgets()
        self.update_selection_textbox(self.selector.get_idxs())

    def update_selection_widgets(self) -> None:
        """Updates the selection widgets so that it matches the full
        setting.
        """
        for check_box in self.check_boxes:
            check_box.blockSignals(True)
            if self.full_settings[check_box.objectName()]:
                check_box.setCheckState(Qt.Checked)
            else:
                check_box.setCheckState(Qt.Unchecked)
            check_box.blockSignals(False)
        for combo_box in self.combo_boxes:
            combo_box.model().blockSignals(True)
            for item in combo_box.getItems():
                txt = item.text()
                if combo_box.objectName() == "index":
                    key = int(txt)
                else:
                    key = txt
                if self.full_settings[combo_box.objectName()][key]:
                    item.setCheckState(Qt.Checked)
                else:
                    item.setCheckState(Qt.Unchecked)
            combo_box.update_all_selected()
            combo_box.update_line_edit()
            combo_box.model().blockSignals(False)

    def update_selection_textbox(self, idxs: set[int]) -> None:
        """Update the selection textbox.

        Parameters
        ----------
        idxs : set[int]
            The selected indexes.
        """
        num_sel = len(idxs)
        text = [f"Number of atoms selected:\n{num_sel}\n\nSelected atoms:\n"]
        atoms = self.selector.system.atom_list
        for idx in idxs:
            text.append(f"{idx}  ({atoms[idx].full_name})\n")
        self.selection_textbox.setText("".join(text))

    def apply(self) -> None:
        """Set the field of the AtomSelectionWidget to the currently
        chosen setting in this widget.
        """
        self.selector.update_settings(self.full_settings)
        self._field.setText(self.selector.settings_to_json())

    def reset(self) -> None:
        """Resets the helper to the default state."""
        self.selector.reset_settings()
        self.selector.settings["all"] = self.all_selection
        self.full_settings = self.selector.full_settings
        self.update_selection_widgets()
        idxs = self.selector.get_idxs()
        self.view_3d._viewer.change_picked(idxs)
        self.update_selection_textbox(idxs)


class AtomSelectionWidget(WidgetBase):
    """The atoms selection widget."""

    _push_button_text = "Atom selection helper"
    _default_value = '{"all": true}'
    _tooltip_text = "Specify which atoms will be used in the analysis. The input is a JSON string, and can be created using the helper dialog."

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = self._default_value
        self._field = QLineEdit(self._default_value, self._base)
        self._field.setPlaceholderText(self._default_value)
        self._field.setMaxLength(2147483647)  # set to the largest possible
        self._field.textChanged.connect(self.updateValue)
        traj_config = self._configurator._configurable[
            self._configurator._dependencies["trajectory"]
        ]
        traj_filename = traj_config["filename"]
        hdf_traj = traj_config["hdf_trajectory"]
        self.helper = self.create_helper((traj_filename, hdf_traj))
        helper_button = QPushButton(self._push_button_text, self._base)
        helper_button.clicked.connect(self.helper_dialog)
        self._layout.addWidget(self._field)
        self._layout.addWidget(helper_button)
        self.update_labels()
        self.updateValue()
        self._field.setToolTip(self._tooltip_text)

    def create_helper(
        self, traj_data: tuple[str, HDFTrajectoryInputData]
    ) -> SelectionHelper:
        """
        Parameters
        ----------
        traj_data : tuple[str, HDFTrajectoryInputData]
            A tuple of the trajectory data used to load the 3D viewer.

        Returns
        -------
        SelectionHelper
            Create and return the selection helper QDialog.
        """
        selector = self._configurator.get_selector()
        return SelectionHelper(selector, traj_data, self._field, self._base)

    @Slot()
    def helper_dialog(self) -> None:
        """Opens the helper dialog."""
        if self.helper.isVisible():
            self.helper.close()
        else:
            self.helper.show()

    def get_widget_value(self) -> str:
        """
        Returns
        -------
        str
            The JSON selector setting.
        """
        selection_string = self._field.text()
        if len(selection_string) < 1:
            self._empty = True
            return self._default_value
        else:
            self._empty = False
        return selection_string



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import Optional
import traceback

import numpy as np
from qtpy.QtWidgets import (
    QPushButton,
    QFileDialog,
    QVBoxLayout,
    QWidget,
    QHBoxLayout,
    QCheckBox,
    QTextEdit,
)
from qtpy.QtCore import Signal, Slot

from MDANSE.MLogging import LOG
from MDANSE.Framework.Jobs.IJob import IJob

from MDANSE_GUI.InputWidgets import *
from MDANSE_GUI.Tabs.Visualisers.InstrumentInfo import SimpleInstrument


widget_lookup = {  # these all come from MDANSE_GUI.InputWidgets
    "FloatConfigurator": FloatWidget,
    "BooleanConfigurator": BooleanWidget,
    "StringConfigurator": StringWidget,
    "IntegerConfigurator": IntegerWidget,
    "CorrelationFramesConfigurator": CorrelationFramesWidget,
    "FramesConfigurator": FramesWidget,
    "RangeConfigurator": RangeWidget,
    "DistHistCutoffConfigurator": DistHistCutoffWidget,
    "VectorConfigurator": VectorWidget,
    "HDFInputFileConfigurator": InputFileWidget,
    "HDFTrajectoryConfigurator": HDFTrajectoryWidget,
    "DerivativeOrderConfigurator": DerivativeOrderWidget,
    "InterpolationOrderConfigurator": InterpolationOrderWidget,
    "OutputFilesConfigurator": OutputFilesWidget,
    "ASEFileConfigurator": InputFileWidget,
    "AseInputFileConfigurator": AseInputFileWidget,
    "ConfigFileConfigurator": InputFileWidget,
    "InputFileConfigurator": InputFileWidget,
    "MDFileConfigurator": InputFileWidget,
    "FieldFileConfigurator": InputFileWidget,
    "XDATCARFileConfigurator": InputFileWidget,
    "XTDFileConfigurator": InputFileWidget,
    "XYZFileConfigurator": InputFileWidget,
    "RunningModeConfigurator": RunningModeWidget,
    "WeightsConfigurator": ComboWidget,
    "MultipleChoicesConfigurator": MultipleCombosWidget,
    "MoleculeSelectionConfigurator": MoleculeWidget,
    "GroupingLevelConfigurator": ComboWidget,
    "SingleChoiceConfigurator": ComboWidget,
    "QVectorsConfigurator": QVectorsWidget,
    "InputDirectoryConfigurator": InputDirectoryWidget,
    "OutputDirectoryConfigurator": OutputDirectoryWidget,
    "OutputStructureConfigurator": OutputStructureWidget,
    "OutputTrajectoryConfigurator": OutputTrajectoryWidget,
    "ProjectionConfigurator": ProjectionWidget,
    "AtomSelectionConfigurator": AtomSelectionWidget,
    "AtomMappingConfigurator": AtomMappingWidget,
    "AtomTransmutationConfigurator": AtomTransmutationWidget,
    "InstrumentResolutionConfigurator": InstrumentResolutionWidget,
    "PartialChargeConfigurator": PartialChargeWidget,
    "UnitCellConfigurator": UnitCellWidget,
}


class Action(QWidget):
    new_thread_objects = Signal(list)
    run_and_load = Signal(list)
    new_path = Signal(str)

    last_paths = {}

    def __init__(self, *args, use_preview=False, **kwargs):
        self._default_path = None
        self._input_trajectory = None
        self._parent_tab = None
        self._trajectory_configurator = None
        self._settings = None
        self._use_preview = use_preview
        self._current_instrument = None
        default_path = kwargs.pop("path", None)
        input_trajectory = kwargs.pop("trajectory", None)
        self.set_trajectory(default_path, input_trajectory)
        super().__init__(*args, **kwargs)

        self.layout = QVBoxLayout(self)
        self.handlers = {}
        self._widgets = []
        self._widgets_in_layout = []

    def set_settings(self, settings):
        self._settings = settings

    def set_trajectory(self, path: Optional[str], trajectory: Optional[str]) -> None:
        """Set the trajectory path and filename.

        Parameters
        ----------
        path : str or None
            The path of the trajectory
        trajectory : str or None
            The path and filename of the trajectory
        """
        self._default_path = path
        self._input_trajectory = trajectory
        path = None
        if self._input_trajectory is not None:
            path, filename = os.path.split(self._input_trajectory)
        if self._default_path is None:
            if path is None:
                self._default_path = "."
            else:
                self._default_path = path

    def set_instrument(self, instrument: SimpleInstrument) -> None:
        self._current_instrument = instrument

    def clear_panel(self) -> None:
        """Clear the widgets so that it leaves an empty layout"""
        for widget in self._widgets_in_layout:
            self.layout.removeWidget(widget)
            # fixes #448
            # even with the call to deleteLater sometimes the widget
            # windows can pop up and then disappear we need to hide
            # them first to make sure this doesn't happen
            widget.hide()
            widget.setParent(None)
            widget.deleteLater()
        self._widgets = []
        self._widgets_in_layout = []
        self._preview_box = None

    def update_panel(self, job_name: str) -> None:
        """Sets all the widgets for the selected job.

        Parameters
        ----------
        job_name : str
            The job name.
        """
        self.clear_panel()

        self._job_name = job_name
        self.last_paths[job_name] = self._parent_tab.get_path(job_name)
        try:
            job_instance = IJob.create(job_name)
        except ValueError as e:
            LOG.debug(
                f"Failed to create IJob {job_name};\n"
                f"reason {e};\n"
                f"traceback {traceback.format_exc()}"
            )
            return
        job_instance.build_configuration()
        settings = job_instance.settings
        self._job_instance = job_instance
        LOG.info(f"Configuration {job_instance.configuration}")
        if "trajectory" in settings.keys():
            if self._input_trajectory is None:
                return
            key, value = "trajectory", settings["trajectory"]
            dtype = value[0]
            ddict = value[1]
            configurator = job_instance.configuration[key]
            if not "label" in ddict.keys():
                ddict["label"] = key
            ddict["configurator"] = configurator
            ddict["source_object"] = self._input_trajectory
            widget_class = widget_lookup[dtype]
            input_widget = widget_class(parent=self, **ddict)
            widget = input_widget._base
            self.layout.addWidget(widget, stretch=input_widget._relative_size)
            self._widgets_in_layout.append(widget)
            self._widgets.append(input_widget)
            self._trajectory_configurator = input_widget._configurator
            LOG.info("Set up input trajectory")
        for key, value in settings.items():
            if key == "trajectory":
                continue
            dtype = value[0]
            ddict = value[1]
            configurator = job_instance.configuration[key]
            if not "label" in ddict.keys():
                ddict["label"] = key
            ddict["configurator"] = configurator
            ddict["source_object"] = self._input_trajectory
            ddict["trajectory_configurator"] = self._trajectory_configurator
            if not dtype in widget_lookup.keys():
                ddict["tooltip"] = (
                    "This is not implemented in the MDANSE GUI at the moment, and it MUST BE!"
                )
                placeholder = BackupWidget(parent=self, **ddict)
                widget = placeholder._base
                self.layout.addWidget(widget, stretch=placeholder._relative_size)
                self._widgets_in_layout.append(widget)
                self._widgets.append(placeholder)
                LOG.warning(f"Could not find the right widget for {key}")
            else:
                widget_class = widget_lookup[dtype]
                # expected = {key: ddict[key] for key in widget_class.__init__.__code__.co_varnames}
                input_widget = widget_class(parent=self, **ddict)
                widget = input_widget._base
                self.layout.addWidget(widget, stretch=input_widget._relative_size)
                self._widgets_in_layout.append(widget)
                self._widgets.append(input_widget)
                input_widget.valid_changed.connect(self.allow_execution)
                if self._use_preview:
                    input_widget.value_updated.connect(self.show_output_prediction)
                LOG.info(f"Set up the right widget for {key}")
            # self.handlers[key] = data_handler
            configured = False
            iterations = 0
            while not configured:
                configured = True
                for widget in self._widgets:
                    widget.value_from_configurator()
                    configured = configured and widget._configurator.is_configured()
                iterations += 1
                if iterations > 5:
                    break

        if self._use_preview:
            self._preview_box = QTextEdit(self)
            self.layout.addWidget(self._preview_box)
            self._widgets_in_layout.append(self._preview_box)

        buttonbase = QWidget(self)
        buttonlayout = QHBoxLayout(buttonbase)
        buttonbase.setLayout(buttonlayout)
        self.save_button = QPushButton("Save as script", buttonbase)
        self.execute_button = QPushButton("RUN!", buttonbase)
        self.execute_button.setStyleSheet("font-weight: bold")
        self.post_execute_checkbox = QCheckBox("Auto-load results", buttonbase)
        try:
            default_check_status = (
                self._parent_tab._settings.group("Execution").get("auto-load") == "True"
            )
        except:
            LOG.debug(f"Converter tab could not load auto-load settings")
            default_check_status = False
        if default_check_status:
            self.post_execute_checkbox.setChecked(True)

        self.save_button.clicked.connect(self.save_dialog)
        self.execute_button.clicked.connect(self.execute_converter)

        buttonlayout.addWidget(self.save_button)
        buttonlayout.addWidget(self.execute_button)
        buttonlayout.addWidget(self.post_execute_checkbox)

        self.layout.addWidget(buttonbase)
        self._widgets_in_layout.append(buttonbase)
        self.apply_instrument()

    def apply_instrument(self):
        if self._current_instrument is not None:
            q_vector_tuple = self._current_instrument.create_q_vector_params()
            resolution_tuple = self._current_instrument.create_resolution_params()
            for widget in self._widgets:
                if isinstance(widget, InstrumentResolutionWidget):
                    if resolution_tuple is None:
                        continue
                    widget.change_function(resolution_tuple[0], resolution_tuple[1])
                if isinstance(widget, QVectorsWidget):
                    if q_vector_tuple is None:
                        continue
                    widget._selector.setCurrentText(q_vector_tuple[0])
                    widget._model.switch_qvector_type(
                        q_vector_tuple[0], q_vector_tuple[1]
                    )
        self.allow_execution()
        self.show_output_prediction()

    @Slot()
    def show_output_prediction(self):
        if self._use_preview:
            self.allow_execution()
            LOG.info("Show output prediction")
            pardict = self.set_parameters()
            self._job_instance.setup(pardict)
            axes = self._job_instance.preview_output_axis()
            LOG.info(f"Axes = {axes.keys()}")
            text = "<p><b>The results will cover the following range:</b></p>"
            for unit, old_array in axes.items():
                scale_factor, new_unit = self._parent_tab.conversion_factor(unit)
                array = np.array(old_array) * scale_factor
                if len(array) < 6:
                    text += f"<p>{array} ({new_unit})</p>"
                else:
                    text += f"<p>[{array[0]}, {array[1]}, {array[2]}, ..., {array[-1]}] ({new_unit})</p>"
            self._preview_box.setHtml(text)

    @Slot(dict)
    def parse_updated_params(self, new_params: dict):
        if "path" in new_params.keys():
            self.default_path = new_params["path"]
            self.new_path.emit(self.default_path)

    @Slot()
    def allow_execution(self):
        allow = True
        for widget in self._widgets:
            if not widget._configurator.valid:
                allow = False
        if allow:
            self.execute_button.setEnabled(True)
        else:
            self.execute_button.setEnabled(False)
        if self._job_name == "AverageStructure":
            self.post_execute_checkbox.setEnabled(False)
        else:
            self.post_execute_checkbox.setEnabled(True)

    @Slot()
    def cancel_dialog(self):
        self.destroy()

    @Slot()
    def save_dialog(self):
        try:
            cname = self._job_name
        except:
            currentpath = "."
        else:
            currentpath = self._parent_tab.get_path(self._job_name + "_script")
        result, ftype = QFileDialog.getSaveFileName(
            self, "Save job as a Python script", currentpath, "Python script (*.py)"
        )
        if result == "":
            return None
        path, _ = os.path.split(result)
        try:
            cname = self._job_name
        except:
            pass
        else:
            self.last_paths[cname] = path
            self._parent_tab.set_path(self._job_name + "_script", path)
        pardict = self.set_parameters(labels=True)
        self._job_instance.save(result, pardict)

    def set_parameters(self, labels=False):
        results = {}
        for widnum, key in enumerate(self._job_instance.settings.keys()):
            if labels:
                label = self._job_instance.settings[key][1]["label"]
                results[key] = (self._widgets[widnum].get_widget_value(), label)
            else:
                results[key] = self._widgets[widnum].get_widget_value()
        return results

    @Slot()
    def execute_converter(self):
        pardict = self.set_parameters()
        LOG.info(pardict)
        self._parent_tab.set_path(self._job_name, self._default_path)
        # when we are ready, we can consider running it
        # self.converter_instance.run(pardict)
        # this would send the actual instance, which _may_ be wrong
        # self.new_thread_objects.emit([self.converter_instance, pardict])
        if (
            self.post_execute_checkbox.isChecked()
            and self._job_name != "AverageStructure"
        ):
            self.run_and_load.emit([self._job_name, pardict])
        else:
            self.new_thread_objects.emit([self._job_name, pardict])



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import os
from functools import partial

from qtpy.QtCore import Slot
from qtpy.QtWidgets import QWidget, QFileDialog

from MDANSE_GUI.Tabs.GeneralTab import GeneralTab
from MDANSE_GUI.Tabs.Layouts.MultiPanel import MultiPanel
from MDANSE_GUI.Session.LocalSession import LocalSession
from MDANSE_GUI.Tabs.Models.PlotDataModel import PlotDataModel
from MDANSE_GUI.Tabs.Views.PlotDataView import PlotDataView
from MDANSE_GUI.Tabs.Visualisers.PlotDataInfo import PlotDataInfo
from MDANSE_GUI.Tabs.Visualisers.DataPlotter import DataPlotter


label_text = """Here you can create plots of specific
data sets. Load the files and assign the data sets
to a plot. The plotting interface will appear
in a new tab of the interface.
"""


class PlotSelectionTab(GeneralTab):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._core.add_button("Load .MDA results", self.load_files)
        self._visualiser._settings = self._settings
        self._visualiser._unit_lookup = self

    def grouped_settings(self):
        results = super().grouped_settings()
        results += [
            [
                "dialogs",
                {"new_plot": "True", "data_plotted": "True", "new_text": "True"},
                {
                    "new_plot": "Show a pop-up dialog EVERY TIME a new plot is created",
                    "data_plotted": "Show a pop-up dialog EVERY TIME a data set is plotted",
                    "new_text": "Show a pop-up dialog EVERY TIME a new data view is created",
                },
            ]
        ]
        return results

    @Slot()
    def load_files(self):
        fnames = QFileDialog.getOpenFileNames(
            self._core,
            "Load an MDA file (MDANSE analysis results)",
            self.get_path("plot_selection"),
            "MDANSE result files (*.mda);;HDF5 files (*.h5);;HDF5 files(*.hdf);;All files(*.*)",
        )
        if fnames is None:
            return
        if len(fnames[0]) < 1:
            return
        for fname in fnames[0]:
            self.load_results(fname)
            last_path, _ = os.path.split(fname)
        self.set_path("plot_selection", last_path)

    @Slot(str)
    def load_results(self, fname: str):
        if len(fname) > 0:
            _, short_name = os.path.split(fname)
            self._model.add_file(fname)

    @classmethod
    def standard_instance(cls):
        the_tab = cls(
            window,
            name="Plotting",
            session=LocalSession(),
            model=PlotDataModel(),
            view=PlotDataView(),
            visualiser=DataPlotter(),
            layout=partial(MultiPanel, left_panels=[PlotDataInfo()]),
            label_text=label_text,
        )
        return the_tab

    @classmethod
    def gui_instance(
        cls,
        parent: QWidget,
        name: str,
        session: LocalSession,
        settings,
        logger,
        **kwargs,
    ):
        the_tab = cls(
            parent,
            name=name,
            session=session,
            settings=settings,
            logger=logger,
            model=kwargs.get("model", PlotDataModel()),
            view=PlotDataView(),
            visualiser=DataPlotter(),
            layout=partial(MultiPanel, left_panels=[PlotDataInfo()]),
            label_text=label_text,
        )
        the_tab._visualiser._unit_lookup = the_tab
        return the_tab


if __name__ == "__main__":
    import sys
    from qtpy.QtWidgets import QApplication, QMainWindow

    app = QApplication(sys.argv)
    window = QMainWindow()
    the_tab = PlotSelectionTab.standard_instance()
    window.setCentralWidget(the_tab._core)
    window.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from abc import abstractmethod

from qtpy.QtCore import QObject, Slot, Signal
from qtpy.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QLabel,
    QGroupBox,
    QVBoxLayout,
    QGridLayout,
)

from MDANSE.MLogging import LOG


class WidgetBase(QObject):

    valid_changed = Signal()
    value_updated = Signal()

    def __init__(self, *args, **kwargs):
        parent = kwargs.get("parent", None)
        super().__init__(*args, parent=parent)
        self._value = None
        self._relative_size = 1
        self._label_text = kwargs.get("label", "")
        self._tooltip = kwargs.pop("tooltip", "")
        self._base_type = kwargs.get("base_type", "QGroupBox")
        self._layout_type = kwargs.get("layout_type", "QHBoxLayout")
        configurator = kwargs.get("configurator", None)
        if self._layout_type == "QHBoxLayout":
            layoutClass = QHBoxLayout
        elif self._layout_type == "QVBoxLayout":
            layoutClass = QVBoxLayout
        else:
            layoutClass = QGridLayout
        if self._base_type == "QWidget":
            base = QWidget(parent)
            layout = layoutClass(base)
            base.setLayout(layout)
            self._label = QLabel(self._label_text, base)
            layout.addWidget(self._label)
        elif self._base_type == "QGroupBox":
            base = QGroupBox(self._label_text, parent)
            base.setToolTip(self._tooltip)
            layout = layoutClass(base)
            base.setLayout(layout)
        self._base = base
        self._base.setObjectName("InputWidget")
        self._layout = layout
        self._configurator = configurator
        self._parent_dialog = parent
        self._empty = False

    def update_labels(self):
        if self._base_type == "QWidget":
            self._label.setText(self._label_text)
        elif self._base_type == "QGroupBox":
            self._base.setTitle(self._label_text)
        for widget in self._layout.children():
            if issubclass(widget, QWidget):
                widget.setToolTip(self._tooltip)

    def default_labels(self):
        """Each Widget should have a default tooltip and label,
        which will be set in this method, unless specific
        values are provided in the settings of the job that
        is being configured."""
        if self._label_text == "":
            self._label_text = "Base Widget"
        if self._tooltip == "":
            self._tooltip = "A specific tooltip for this widget SHOULD have been set"

    @abstractmethod
    def value_from_configurator(self):
        """Set the widgets to the values of the underlying configurator object.
        Should also check for dependencies of the configurator.
        """

    @abstractmethod
    def get_widget_value(self):
        """Collect the results from the input widgets and return the value."""

    @abstractmethod
    def configure_using_default(self):
        """Makes the configurator use its default value, and highlights it
        in the GUI"""
        default = self._configurator.default
        LOG.info(f"Setting {default} as placeholder text")
        self._field.setPlaceholderText(str(default))
        self._configurator.configure(default)

    def mark_error(self, error_text: str):
        self._base.setStyleSheet(
            "QWidget#InputWidget { background-color:rgb(180,20,180); font-weight: bold }"
        )
        self._base.setToolTip(error_text)
        self.valid_changed.emit()

    def clear_error(self):
        self._base.setStyleSheet("")
        self._base.setToolTip("")
        self.valid_changed.emit()

    @abstractmethod
    @Slot()
    def updateValue(self):
        current_value = self.get_widget_value()
        if self._empty:
            self.configure_using_default()
        try:
            self._configurator.configure(current_value)
        except:
            self.mark_error(
                "COULD NOT SET THIS VALUE - you may need to change the values in other widgets"
            )
        if self._configurator.valid:
            self.clear_error()
            self.value_updated.emit()
        else:
            self.mark_error(self._configurator.error_status)

    @abstractmethod
    def get_value(self):
        self.updateValue()
        return self._configurator["value"]

    @property
    def default_path(self):
        return self._parent_dialog.default_path

    @default_path.setter
    def default_path(self, value: str) -> None:
        self._parent_dialog.default_path = value



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import Union, Iterable

import copy

from qtpy.QtWidgets import (
    QDialog,
    QPushButton,
    QFileDialog,
    QVBoxLayout,
    QWidget,
    QLabel,
    QComboBox,
    QLineEdit,
    QFormLayout,
    QHBoxLayout,
    QCheckBox,
)
from qtpy.QtCore import Signal, Slot, QObject
from qtpy.QtGui import (
    QIntValidator,
    QDoubleValidator,
)

from MDANSE.MLogging import LOG

# I think that a Trajectory Converter should, in general,
# create a Wizard and not a single Dialog.
# This way LAMMPS could give the user a chance to verify which
# elements were in the system, instead of just failing.
# The idea is similar to Origin and the way Origin creates
# a lengthy wizard interface just to load data from a file.
# The wizard should have a number of stages, which can be
# used or left empty.

# I need to have several stages of processing
# 1. The initial Dialog is created based on settings from the Converter
# 2. The user specifies the input files
# 3. The converter tries to get as much information as possible from the files
# 4. The information gathered is shown to the user,
# with a possibility of correcting the entries.
# 5. The corrected parameters are passed to the converter, and the job is started.


class GeneralInput(QObject):
    """
    This is a standard object for storing information from input fields.
    It inherits QObject, and so it is able to use the signal/slot mechanism.
    For any input variable, the idea is to attach a GeneralInput to the
    corresponding GUI element to allow the software to handle all the
    data flow in a standardised way.
    """

    final_value = Signal(object)
    string_value = Signal(str)
    value_changed = Signal()

    def __init__(self, *args, data_type=None, **kwargs):
        new_kwargs = copy.copy(kwargs)
        for kkey in InputFactory.reserved_keywords:
            if kkey in new_kwargs.keys():
                new_kwargs.pop(kkey)
        super().__init__(*args, **{})

        self.default_path = kwargs.get("path", None)
        self.data_type = data_type
        self.default_value = kwargs.get("default", None)
        self.current_value = self.default_value
        self.file_association = kwargs.get("file_association", None)
        self.file_direction = kwargs.get("file_direction", "none")
        self.file_format = kwargs.get("format", "")
        if self.file_association is not None:
            if len(self.file_association) < 1 and len(self.file_format) > 1:
                self.file_association = "*." + self.file_format
        if self.file_direction == "in":
            self.file_dialog = QFileDialog.getOpenFileName
        elif self.file_direction == "out":
            self.file_dialog = QFileDialog.getSaveFileName
        else:
            self.file_dialog = None
        LOG.info(kwargs)

    @Slot(str)
    def updatePath(self, newpath: str):
        """Interesting only to the variables using a FileDialog,
        this slot allows the Dialog to update the default path
        to be used for finding the input/output files.

        Arguments:
            newpath -- the new path in the filesystem to be used
            in new instances of FileDialog.
        """
        self.default_path = newpath

    @Slot(str)
    @Slot(int)
    @Slot(object)
    def updateValue(self, newone, emit=False):
        """This slot allows the GeneralInput to receive a new value
        from the corresponding GUI element. In principle, the GUI
        should have produced a valid input, but tests for valid
        conversion are still being performed here.

        Arguments:
            newone -- new value to be stored by the GeneralInput instance

        Keyword Arguments:
            emit -- if True, the GeneralInput instance will broadcast
            the new value as a signal.
        """
        try:
            converted = self.data_type(newone)
        except ValueError:
            converted = self.default_value
            LOG.error(f"ValueError converting {newone} using {self.data_type}")
        except TypeError:
            converted = self.default_value
            LOG.error(f"TypeError converting {newone} using {self.data_type}")
        self.current_value = converted
        self.value_changed.emit()
        if emit:
            self.string_value.emit(self.current_value)

    def returnValue(self):
        """Returns the value stored by this instance.
        Typically used when the Dialog has been filled out,
        and the inputs have to be summarised.

        Returns:
            The value revceived from the GUI, in the format
            that was specified in the constructor.
        """
        if self.file_dialog is not None:
            LOG.info(f"File Field Return Value: {self.current_value}")
        self.final_value.emit(self.current_value)
        return self.current_value

    @Slot()
    def valueFromDialog(self):
        """A Slot defined to allow the GUI to be updated based on
        the new path received from a FileDialog.
        This will start a FileDialog, take the resulting path,
        and emit a signal to update the value show by the GUI.
        """
        new_value = self.file_dialog(
            self.parent(),  # the parent of the dialog
            "Load a file",  # the label of the window
            self.default_path,  # the initial search path
            self.file_association,  # text string specifying the file name filter.
        )
        if new_value is not None:
            self.updateValue(new_value[0], emit=True)


# class MultipleInput(GeneralInput):
#     """
#     This class is more specifically designed for handling variables
#     with multiple values of the same type.
#     """

#     def __init__(self, *args, data_type=None, **kwargs):
#         super().__init__(*args, data_type=data_type, **kwargs)
#         self.length = kwargs.get("length", 1)
#         if isinstance(self.default_value, Iterable) and not isinstance(self.default_value, str):
#             if len(self.default_value) == self.length:
#                 self.default_value = np.array(list(self.default_value), dtype=self.data_type)
#             elif len(self.default_value) > self.length:
#                 self.default_value = np.array(list(self.default_value)[:self.length], dtype=self.data_type)
#             else:
#                 padding = self.length - len(self.default_value)
#                 self.default_value = np.array(list(self.default_value) + padding*[0]).astype(self.data_type)
#         else:
#             self.default_value = np.array(self.length*[self.default_value], dtype=self.data_type)
#         self.current_value = self.default_value


class InputGroup(QObject):
    final_value = Signal(object)
    string_value = Signal(str)

    def __init__(self, parent: Union[QObject, None] = ...) -> None:
        super().__init__(parent)
        self.fields = []
        self.values = []

    @Slot()
    def valueChanged(self):
        """This slot is triggered when one of the inputs
        changes its value. It then checks ALL the inputs
        and outputs their values put together."""
        result = []
        for field in self.fields:
            result.append(field.returnValue())
        self.final_value.emit(result)
        self.string_value.emit(str(result))

    def register_value(self, field: GeneralInput):
        """Adds an input field to the list of fields"""
        self.fields.append(field)
        field.value_changed.connect(self.valueChanged)

    def returnValue(self):
        result = []
        for field in self.fields:
            result.append(field.returnValue())
        return result


class InputFactory:
    """Factory generating input fields for Dialogs. It is meant to standardise
    the way inputs are hanlded, so that any kind of variable can be processed
    using the same methods.
    """

    reserved_keywords = [
        "kind",
        "default",
        "label",
        "tooltip",
        "wildcard",
        "mini",
        "choices",
        "file_association",
        "file_direction",
    ]

    def createInputField(*args, **kwargs):
        """Creates an input field for the specified kind of variable.
        The returned base widget is a QWidget with its own layout and
        children, which should be placed in the layout of the Dialog.
        The variables themselves are passed to the Dialog by the
        GeneralInput object, to make the exchange independent of the
        type of widgets used to construct the input field.

        Returns:
            [QWidget, GeneralInput] pair
        """
        kind = kwargs.get("kind", "String")

        if kind == "StringConfigurator":
            result = InputFactory.createSingle(*args, **kwargs)
        elif kind == "PythonObjectConfigurator":
            result = InputFactory.createSingle(*args, **kwargs)
        elif kind == "IntegerConfigurator":
            result = InputFactory.createSingle(*args, **kwargs)
        elif kind == "RangeConfigurator":
            result = InputFactory.createMultiple(*args, **kwargs)
        elif kind == "FloatConfigurator":
            result = InputFactory.createSingle(*args, **kwargs)
        elif kind == "SingleChoiceConfigurator":
            result = InputFactory.createCombo(*args, **kwargs)
        elif kind == "BooleanConfigurator":
            result = InputFactory.createBool(*args, **kwargs)
        elif kind == "InputFileConfigurator":
            result = InputFactory.createFile(*args, direction="in", **kwargs)
        elif kind == "SingleOutputFileConfigurator":
            result = InputFactory.createFile(*args, direction="out", **kwargs)
        elif kind == "OutputTrajectoryConfigurator":
            result = InputFactory.createTrajectory(*args, direction="out", **kwargs)
        else:
            result = InputFactory.createBlank(*args, **kwargs)

        return result

    def createBase(*args, **kwargs):
        """Some parts of the input will always be the same,
        so we handle them in one function that will be called
        before we get to the specific details

        Returns:
            [base, layout] pair of QWidget and associated layout
        """
        parent = kwargs.get("parent", None)
        label_text = kwargs.get("label", None)
        tooltip_text = kwargs.get("tooltip", None)
        base = QWidget(parent)
        layout = QHBoxLayout(base)
        base.setLayout(layout)
        label = QLabel(label_text, base)
        label.setToolTip(tooltip_text)
        layout.addWidget(label)
        return [base, layout]

    def createBlank(*args, **kwargs):
        """This method creates a placeholder which will inform
        the users that the requested input field could not
        be constructed.
        """
        kind = kwargs.get("kind", "String")
        base, layout = InputFactory.createBase(
            label=f"<b>MISSING TYPE</b>:{kind}",
            tooltip="This is not handled by the MDANSE GUI correctly! Please report the problem to the authors.",
        )
        return [base, GeneralInput()]

    def createFile(*args, direction="in", **kwargs):
        """Creates an input field with an additional button which
        creates a FileDialog, to allow the user to browse the file system.

        Keyword Arguments:
            direction -- if 'in', create a FileDialog for an exisitng file,
              if 'out', a FileDialog for creating a new file.
        """
        kind = kwargs.get("kind", "String")
        default_value = kwargs.get("default", "")
        tooltip_text = kwargs.get("tooltip", "Specify a path to an existing file.")
        qt_file_association = file_association = kwargs.get("wildcard", "")
        LOG.info(kind)
        LOG.info(default_value)
        LOG.info(tooltip_text)
        LOG.info(file_association)
        LOG.info(kwargs)
        base, layout = InputFactory.createBase(*args, **kwargs)
        field = QLineEdit(base)
        data_handler = GeneralInput(
            data_type=str,
            file_association=qt_file_association,
            file_direction=direction,
            **kwargs,
        )
        data_handler.string_value.connect(field.setText)
        field.textChanged.connect(data_handler.updateValue)
        field.setText(str(default_value))
        field.setToolTip(tooltip_text)
        layout.addWidget(field)
        button = QPushButton("Browse", base)
        button.clicked.connect(data_handler.valueFromDialog)
        layout.addWidget(button)
        return [base, data_handler]

    def createTrajectory(*args, direction="out", **kwargs):
        """Creates an input field with an additional button which
        creates a FileDialog, to allow the user to browse the file system.

        Keyword Arguments:
            direction -- if 'in', create a FileDialog for an exisitng file,
              if 'out', a FileDialog for creating a new file.
        """
        kind = kwargs.get("kind", "String")
        default_value = kwargs.get("default", "")
        tooltip_text = kwargs.get(
            "tooltip", "Specify the name of the output trajectory."
        )
        qt_file_association = kwargs.get("wildcard", "")
        base, layout = InputFactory.createBase(*args, **kwargs)
        field = QLineEdit(base)
        main_handler = InputGroup(base)
        data_handler1 = GeneralInput(
            data_type=str,
            file_association=qt_file_association,
            file_direction=direction,
            **kwargs,
        )
        data_handler2 = GeneralInput(
            data_type=int,
        )
        data_handler3 = GeneralInput(
            data_type=str,
        )
        main_handler.register_value(data_handler1)
        main_handler.register_value(data_handler2)
        main_handler.register_value(data_handler3)
        data_handler1.string_value.connect(field.setText)
        field.textChanged.connect(data_handler1.updateValue)
        field.setText(str(default_value))
        field.setToolTip(tooltip_text)
        layout.addWidget(field)
        cbox1 = QComboBox(base)
        cbox1.addItems(["16", "32", "64"])
        cbox1.setCurrentText("64")
        layout.addWidget(cbox1)
        cbox2 = QComboBox(base)
        cbox2.addItems(["no compression", "gzip"])
        cbox2.setCurrentText("gzip")
        layout.addWidget(cbox2)
        cbox1.currentTextChanged.connect(data_handler2.updateValue)
        cbox2.currentTextChanged.connect(data_handler3.updateValue)
        data_handler2.updateValue("64")
        data_handler3.updateValue("gzip")
        button = QPushButton("Browse", base)
        button.clicked.connect(data_handler1.valueFromDialog)
        layout.addWidget(button)
        return [base, main_handler]

    def createBool(*args, **kwargs):
        """Creates an input field for a logical variable,
        which is currently implemented as a check box.
        """
        kind = kwargs.get("kind", "Boolean")
        default_value = kwargs.get("default", False)
        tooltip_text = kwargs.get("tooltip", None)
        LOG.info(kind)
        LOG.info(default_value)
        LOG.info(tooltip_text)
        LOG.info(kwargs)
        base, layout = InputFactory.createBase(*args, **kwargs)
        field = QCheckBox(base)
        field.setTristate(False)
        data_handler = GeneralInput(data_type=bool, **kwargs)
        field.stateChanged.connect(data_handler.updateValue)
        field.setChecked(bool(default_value))
        field.setToolTip(tooltip_text)
        layout.addWidget(field)
        return [base, data_handler]

    def createSingle(*args, **kwargs):
        """Creates a LineEdit input for a single variable.
        Can be used to handle int, float or str variables.
        For numerical values, it adds a QValidator instance
        to filter out invalid inputs.
        """
        kind = kwargs.get("kind", "str")
        default_value = kwargs.get("default", None)
        tooltip_text = kwargs.get("tooltip", None)
        minval = kwargs.get("mini", None)
        LOG.info(kind)
        LOG.info(default_value)
        LOG.info(tooltip_text)
        LOG.info(kwargs)
        base, layout = InputFactory.createBase(*args, **kwargs)
        field = QLineEdit(base)
        if "Integer" in kind:
            data_handler = GeneralInput(data_type=int, **kwargs)
            validator = QIntValidator(field)
        elif "Float" in kind:
            data_handler = GeneralInput(data_type=float, **kwargs)
            validator = QDoubleValidator(field)
        else:
            data_handler = GeneralInput(data_type=str, **kwargs)
            validator = None
        if validator is not None:
            field.setValidator(validator)
            if minval is not None:
                validator.setBottom(minval)
        field.textChanged.connect(data_handler.updateValue)
        field.setText(str(default_value))
        field.setToolTip(tooltip_text)
        layout.addWidget(field)
        return [base, data_handler]

    def createMultiple(*args, **kwargs):
        """Creates a number of LineEdit fields, for
        input of multiple values.
        Can be used to handle int, float or str variables.
        For numerical values, it adds a QValidator instance
        to filter out invalid inputs.
        """
        kind = kwargs.get("kind", "String")
        default_value = kwargs.get("default", None)
        tooltip_text = kwargs.get("tooltip", None)
        minval = kwargs.get("mini", None)
        number_of_fields = kwargs.get("howmany", None)
        if number_of_fields is None:
            number_of_fields = len(default_value)
        LOG.info(kind)
        LOG.info(default_value)
        LOG.info(tooltip_text)
        LOG.info(kwargs)
        base, layout = InputFactory.createBase(*args, **kwargs)
        main_handler = InputGroup(base)
        for nfield in range(number_of_fields):
            field = QLineEdit(base)
            if "Integer" in kind:
                data_handler = GeneralInput(data_type=int, **kwargs)
                validator = QIntValidator(field)
            elif "Float" in kind:
                data_handler = GeneralInput(data_type=float, **kwargs)
                validator = QDoubleValidator(field)
            else:
                data_handler = GeneralInput(data_type=str, **kwargs)
                validator = None
            if validator is not None:
                field.setValidator(validator)
                if minval is not None:
                    validator.setBottom(minval)
            field.textChanged.connect(data_handler.updateValue)
            field.setText(str(default_value[nfield]))
            field.setToolTip(tooltip_text)
            main_handler.register_value(data_handler)
        layout.addWidget(field)
        return [base, main_handler]

    def createCombo(*args, **kwargs):
        """For the variable where one option has to be picked from
        a list of possible values, we create a ComboBox"""
        kind = kwargs.get("kind", "String")
        default_value = kwargs.get("default", False)
        tooltip_text = kwargs.get("tooltip", None)
        option_list = kwargs.get("choices", [])
        LOG.info(kind)
        LOG.info(default_value)
        LOG.info(tooltip_text)
        LOG.info(kwargs)
        base, layout = InputFactory.createBase(*args, **kwargs)
        field = QComboBox(base)
        field.addItems(option_list)
        data_handler = GeneralInput(data_type=bool, **kwargs)
        field.currentTextChanged.connect(data_handler.updateValue)
        field.setToolTip(tooltip_text)
        layout.addWidget(field)
        return [base, data_handler]


# the following code was useful at first - but could it be replaced by
# the widgets and dialogs above?
# The InputVariable/InputDialog combo seems to be offering a subset
# of what InputFactory and GeneralInput can do.


class InputVariable(QObject):
    """A general-purpose input field, used by the InputDialog."""

    def __init__(self, *args, input_dict: dict = None, **kwargs):
        super().__init__(*args, **kwargs)

        self.label = "Variable"
        self.keyval = "var1"
        self.format = float
        self.values = [0.0]
        self.widget = QLineEdit
        self.number_of_inputs = 1
        self.helper_dialog = None
        self.tooltip = ""
        self.placeholder = ""
        for key, item in input_dict.items():
            self.__setattr__(str(key), item)

        self.input_widgets = []

    def returnValues(self):
        result = []
        for widget in self.input_widgets:
            try:
                temp = self.format(widget.text())
            except ValueError:
                temp = widget.text()
            result.append(temp)
        return result


class InputDialog(QDialog):
    """A simple dialog designed to contain a number of input fields.
    It is used by the UnitsEditor and ElementsDatabaseEditor.
    """

    got_values = Signal(dict)

    def __init__(self, *args, fields: Iterable["InputVariable"] = None, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields = fields  #  we need to store it for later
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        # now some basic elements
        var_base = QWidget(self)
        var_layout = QFormLayout(var_base)
        var_base.setLayout(var_layout)
        button_base = QWidget(self)
        button_layout = QHBoxLayout(button_base)
        button_base.setLayout(button_layout)
        layout.addWidget(var_base)
        layout.addWidget(button_base)
        # the most important part: handling the variables
        for var in fields:
            label = var.label
            format = var.format
            values = var.values
            widget = var.widget
            helper_dialog = var.helper_dialog
            tooltip = var.tooltip
            placeholder = var.placeholder
            number_of_inputs = var.number_of_inputs
            # set up widgets
            temp_base = QWidget(var_base)
            temp_layout = QHBoxLayout(temp_base)
            temp_base.setLayout(temp_layout)
            if format is int:
                validator = QIntValidator(temp_base)
            elif format is float:
                validator = QDoubleValidator(temp_base)
            else:
                validator = None
            for ni in range(number_of_inputs):
                widget_instance = widget(var_base)
                widget_instance.setText(str(values[ni]))
                widget_instance.setToolTip(tooltip)
                widget_instance.setPlaceholderText(placeholder)
                if validator is not None:
                    widget_instance.setValidator(validator)
                temp_layout.addWidget(widget_instance)
                var.input_widgets.append(widget_instance)
            var_layout.addRow(label, temp_base)
        # optinally we can add other buttons here...
        #
        # final button, always there
        self.button = QPushButton("Accept!", self)
        self.button.clicked.connect(self.accept)
        self.accepted.connect(self.return_value)
        button_layout.addWidget(self.button)

    @Slot()
    def return_value(self):
        result = {}
        for var in self.fields:
            temp = var.returnValues()
            if var.number_of_inputs == 1:
                value = temp[0]
            else:
                value = temp
            result[var.keyval] = value
        self.got_values.emit(result)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#


from typing import Dict, Tuple

from qtpy.QtCore import QObject, Slot, QMessageLogger
from qtpy.QtWidgets import QListView

from MDANSE.MLogging import LOG
from MDANSE.Framework.Units import measure, unit_lookup

from MDANSE_GUI.Tabs.Layouts.DoublePanel import DoublePanel
from MDANSE_GUI.Session.LocalSession import LocalSession
from MDANSE_GUI.Tabs.Visualisers.TextInfo import TextInfo


class GeneralTab(QObject):
    """This object connects different elements of a GUI tab,
    such as the data model, view, visualised, layout,
    session, settings and project, all of them relevant
    to the MDANSE_GUI design.

    The idea of tying the well-defined GUI elements into
    a fairly abstract concept of a 'general tab' is intended
    to give the programmers enough flexibility to change the
    behaviour of GUI sections while keeping the common API
    for accessing them from the outside.
    """

    def __init__(self, *args, **kwargs):
        self._name = kwargs.pop("name", "Unnamed GUI part")
        self._session = kwargs.pop("session", LocalSession())
        _ = kwargs.pop("settings", None)
        self._settings = self._session.obtain_settings(self)
        try:
            self._global_settings = self._session.main_settings()
        except AttributeError:
            self._global_settings = None
        self._model = kwargs.pop("model", None)
        self._visualiser = kwargs.pop("visualiser", TextInfo())
        self._view = kwargs.pop("view", QListView())
        self._logger = kwargs.pop("logger", QMessageLogger())
        layout = kwargs.pop("layout", DoublePanel)
        label_text = kwargs.pop("label_text", "An abstract GUI element")
        super().__init__(*args, **kwargs)
        self._core = layout(
            **{
                "data_side": self._view,
                "visualiser_side": self._visualiser,
                "tab_reference": self,
            }
        )
        if self._model is not None:
            self._core.set_model(self._model)
        self._core.set_label_text(label_text)
        self.propagate_session()

    def grouped_settings(self):
        """This method tells the Session object what settings
        this Tab will store in its settings file,
        and what each of the settings means.

        This way each GUI component can have its own settings
        without interfering with the others, and the user
        does not risk breaking all the settings by
        mangling a single file.

        Returns
        -------
        List[str, Dict[str, str], Dict[str, str]]
        """
        group1 = [
            "Generic settings",  # name of the group of settings
            {"path": "."},  # a dictionary of settings
            {
                "path": "The path last used by this GUI element."
            },  # a dictionary of comments
        ]
        group2 = [
            "units",  # name of the group of settings
            {"energy": "meV", "time": "fs", "distance": "ang", "reciprocal": "1/ang"},
            {
                "energy": "The unit of energy preferred by the user.",
                "time": "The unit of time preferred by the user.",
                "distance": "The unit of distance preferred by the user",
                "reciprocal": "The momentum (transfer) unit preferred by the user",
            },
        ]
        return [group1, group2]

    def connect_units(self):
        if self._visualiser is not None:
            if self._visualiser._unit_lookup is None:
                LOG.debug(f"Visualiser {self._visualiser} has no unit lookup")
                self._visualiser._unit_lookup = self

    def conversion_factor(self, input_unit: str) -> Tuple[float, str]:
        """Finds the conversion factor from an input unit
        to the unit preferred by the user for a given
        physical property.

        Parameters
        ----------
        input_unit : str
            Name/abbreviation of a physical unit

        Returns
        -------
        Tuple[float, str]
            factor F and text label str
            Conversion factor F for converting from the input unit
            to the unit saved by the LocalSession instance.
            The conversion will be done outside of this
            function, following the formula:
            converted_value = F * input_value
        """
        conversion_factor = 1.0
        target_unit = input_unit
        property = unit_lookup.get(input_unit, "unknown")
        unit_group = self._settings.group("units").as_dict()
        backup_group = self._global_settings.group("units").as_dict()
        if property not in unit_group:
            if property not in backup_group:
                return conversion_factor, target_unit
            else:
                target_unit = backup_group[property]
        else:
            target_unit = unit_group[property]
        try:
            conversion_factor = measure(1.0, input_unit, equivalent=True).toval(
                target_unit
            )
        except:
            target_unit = self._settings.default_value("units", property)
            conversion_factor = measure(1.0, input_unit, equivalent=True).toval(
                target_unit
            )
        return conversion_factor, target_unit

    def get_path(self, path_key: str):
        paths_group = self._settings.group("paths")
        try:
            path = paths_group.get(path_key)
        except KeyError:
            paths_group.add(
                path_key, ".", f"Filesystem path recently used by {path_key}"
            )
            path = "."
        return path

    def set_path(self, path_key: str, path_value: str):
        paths_group = self._settings.group("paths")
        if not paths_group.set(path_key, path_value):
            paths_group.add(
                path_key, path_value, f"Filesystem path recently used by {path_key}"
            )

    @Slot()
    def save_state(self):
        self._session.save_state(self)

    def load_state(self):
        self._session.load_state(self)

    def propagate_session(self):
        for target in [self._model, self._visualiser, self._view, self._logger]:
            if target is not None:
                target._session = self._session

    @Slot(str)
    def critical(self, message: str):
        LOG.critical(message)

    @Slot(str)
    def warning(self, message: str):
        LOG.warning(message)

    @Slot(str)
    def error(self, message: str):
        LOG.error(message)

    @Slot(str)
    def debug(self, message: str):
        LOG.debug(message)

    @Slot(str)
    def info(self, message: str):
        LOG.info(message)

    @Slot(str)
    def fatal(self, message: str):
        LOG.critical(message)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import os

from qtpy.QtWidgets import QLineEdit, QPushButton, QFileDialog
from qtpy.QtCore import Slot

from MDANSE.MLogging import LOG

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class InputFileWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        configurator = kwargs.get("configurator", None)
        if configurator is not None:
            default_value = configurator.default
        else:
            default_value = ""
        parent = kwargs.get("parent", None)
        if parent is not None:
            self._job_name = parent._job_name
            self._settings = parent._settings
        try:
            paths_group = self._settings.group("paths")
            try:
                self.default_path = paths_group.get(self._job_name)
            except KeyError:
                paths_group.add(
                    self._job_name,
                    ".",
                    f"The filesystem path recently used by {self._job_name}",
                )
                self.default_path = "."
        except AttributeError:
            try:
                self.default_path = parent.default_path
            except KeyError:
                self.default_path = "."
                LOG.error("KeyError in InputFileWidget - can't get default path.")
            except AttributeError:
                self.default_path = "."
                LOG.error("AttributeError in InputFileWidget - can't get default path.")
        default_value = kwargs.get("default", "")
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "Specify a path to an existing file."
        try:
            file_association = configurator.wildcard
        except AttributeError:
            file_association = kwargs.get("wildcard", "")
        self._qt_file_association = file_association
        field = QLineEdit(self._base)
        self._field = field
        field.textChanged.connect(self.updateValue)
        field.setText(str(default_value))
        field.setPlaceholderText(str(default_value))
        field.setToolTip(tooltip_text)
        self._layout.addWidget(field)
        button = QPushButton("Browse", self._base)
        button.clicked.connect(self.valueFromDialog)
        self._default_value = default_value
        self._layout.addWidget(button)
        self._configurator = configurator
        self._file_dialog = QFileDialog.getOpenFileName
        self.updateValue()

    def configure_using_default(self):
        """This is too specific to have a default value"""

    @Slot()
    def valueFromDialog(self):
        """A Slot defined to allow the GUI to be updated based on
        the new path received from a FileDialog.
        This will start a FileDialog, take the resulting path,
        and emit a signal to update the value show by the GUI.
        """
        paths_group = self._settings.group("paths")
        try:
            self.default_path = paths_group.get(self._job_name)
        except:
            LOG.warning(f"session.get_path failed for {self._job_name}")
        new_value = self._file_dialog(
            self.parent(),  # the parent of the dialog
            "Load file",  # the label of the window
            self.default_path,  # the initial search path
            self._qt_file_association,  # text string specifying the file name filter.
        )
        if new_value is not None:
            if new_value[0]:
                self._field.setText(new_value[0])
                self.updateValue()
                try:
                    LOG.info(
                        f"Settings path of {self._job_name} to {os.path.split(new_value[0])[0]}"
                    )
                    paths_group.set(self._job_name, os.path.split(new_value[0])[0])
                except:
                    LOG.error(
                        f"session.set_path failed for {self._job_name}, {os.path.split(new_value[0])[0]}"
                    )

    def get_widget_value(self):
        """Collect the results from the input widgets and return the value."""
        strval = self._field.text()
        if len(strval) < 1:
            self._empty = True
            return self._default_value
        else:
            self._empty = False
        return strval



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QLineEdit, QLabel
from qtpy.QtGui import QDoubleValidator, QIntValidator

from MDANSE.MLogging import LOG

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class RangeWidget(WidgetBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, layout_type="QGridLayout", **kwargs)
        self.setup_fields(*args, **kwargs)

    def setup_fields(self, *args, **kwargs):
        start, end, step = kwargs.get("default", self._configurator._default)
        num_type = kwargs.get("valueType", int)

        labels = [
            QLabel("From", self._base),
            QLabel("to", self._base),
            QLabel("in steps of", self._base),
        ]
        fields = [
            QLineEdit(str(num_type(start)), self._base),
            QLineEdit(str(num_type(end)), self._base),
            QLineEdit(str(num_type(step)), self._base),
        ]
        placeholders = [str(num_type(start)), str(num_type(end)), str(num_type(step))]
        if num_type is int:
            validators = [QIntValidator(parent_field) for parent_field in fields]
        else:
            validators = [QDoubleValidator(parent_field) for parent_field in fields]
        for field_num in range(3):
            self._layout.addWidget(labels[field_num], 0, 2 * field_num)
            self._layout.addWidget(fields[field_num], 0, 2 * field_num + 1)
            fields[field_num].setValidator(validators[field_num])
            fields[field_num].textChanged.connect(self.updateValue)
            fields[field_num].setPlaceholderText(placeholders[field_num])
        self._fields = fields
        self._validators = validators
        self._num_type = num_type
        self._default_values = placeholders
        self.default_labels()
        self.update_labels()
        self.updateValue()
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "Values to be used, given as (First, Last, StepSize)"
        for wid in fields + labels:
            wid.setToolTip(tooltip_text)

    def configure_using_default(self):
        """This is too complex to have a default value"""

    def default_labels(self):
        """Each Widget should have a default tooltip and label,
        which will be set in this method, unless specific
        values are provided in the settings of the job that
        is being configured."""
        if self._label_text == "":
            self._label_text = "RangeWidget"
        if self._tooltip == "":
            self._tooltip = "Values to be used, given as (First, Last, StepSize)"

    def value_from_configurator(self):
        if self._configurator.check_dependencies():
            minval, maxval = self._configurator._mini, self._configurator._maxi
            LOG.info(f"Configurator min/max: {minval}, {maxval}")
            if maxval is None:
                return
            for val in self._validators:
                val.setBottom(-abs(maxval))
                val.setTop(abs(maxval))

    def get_widget_value(self):
        result = []
        for n, field in enumerate(self._fields):
            strval = field.text()
            try:
                val = self._num_type(strval)
            except:
                val = self._num_type(self._default_values[n])
            result.append(val)
        return result



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from qtpy.QtWidgets import QTreeView
from qtpy.QtCore import Signal

from MDANSE_GUI.Tabs.Views.Delegates import ColourPicker, MplStyleCombo, MainAxisCombo
from MDANSE_GUI.Tabs.Models.PlottingContext import (
    get_mpl_lines,
    get_mpl_markers,
    plotting_column_index,
)


class PlotDetailsView(QTreeView):
    details_changed = Signal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._delegates = {
            "colour": ColourPicker(),
            "axis": MainAxisCombo(),
            "line": MplStyleCombo(mpl_items=get_mpl_lines()),
            "marker": MplStyleCombo(mpl_items=get_mpl_markers()),
        }
        self.setItemDelegateForColumn(
            plotting_column_index["Colour"], self._delegates["colour"]
        )
        self.setItemDelegateForColumn(
            plotting_column_index["Line style"], self._delegates["line"]
        )
        self.setItemDelegateForColumn(
            plotting_column_index["Marker"], self._delegates["marker"]
        )
        self.setItemDelegateForColumn(
            plotting_column_index["Main axis"], self._delegates["axis"]
        )

    def connect_to_visualiser(self, visualiser) -> None:
        """Connect to a visualiser.

        Parameters
        ----------
        visualiser : Action or TextInfo
            A visualiser to connect to this view.
        """
        self.details_changed.connect(visualiser.update_plot_details)



import pytest
import tempfile
import os

import h5py
import numpy as np

from qtpy import QtGui, QtCore, QtWidgets

from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext, SingleDataset

file_1d_name = "super_dos.mda"
file_2d_name = "disf_argon.mda"


@pytest.fixture()
def file_1d():
    tempfile = h5py.File(file_1d_name)
    yield tempfile
    tempfile.close()


@pytest.fixture()
def file_2d():
    tempfile = h5py.File(file_2d_name)
    yield tempfile
    tempfile.close()


def test_single_dataset_1d(file_1d):
    temp = SingleDataset("dos_total", file_1d)
    assert len(temp._data.shape) == 1


def test_single_dataset_2d(file_2d):
    temp = SingleDataset("f(q,t)_total", file_2d)
    assert len(temp._data.shape) == 2


def test_available_x_axes_1d(file_1d):
    temp = SingleDataset("dos_total", file_1d)
    axes = temp.available_x_axes()
    assert axes == ["rad/ps"]
    longest = temp.longest_axis()
    assert longest[0] == "rad/ps"


def test_available_x_axes_2d(file_2d):
    temp = SingleDataset("f(q,t)_total", file_2d)
    axes = temp.available_x_axes()
    print(axes)
    assert axes == ["1/nm", "ps"]
    longest = temp.longest_axis()
    print(longest)
    assert longest[0] == "ps"


def test_curves_vs_axis_1d(file_1d):
    temp = SingleDataset("dos_total", file_1d)
    curves = temp.curves_vs_axis("rad/ps")
    assert len(curves) == 1
    print(curves)
    assert len(curves[tuple()]) == 501


def test_curves_vs_axis_2d_long_axis(file_2d):
    temp = SingleDataset("f(q,t)_total", file_2d)
    curves = temp.curves_vs_axis("ps")
    print(len(curves))
    assert len(curves) == 10
    print(curves.keys())
    assert len(curves[(0,)]) == 501


def test_curves_vs_axis_2d_short_axis(file_2d):
    temp = SingleDataset("f(q,t)_total", file_2d)
    curves = temp.curves_vs_axis("1/nm")
    print(len(curves))
    assert len(curves) == 501
    print(curves.keys())
    assert len(curves[(0,)]) == 10



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import json
from collections import defaultdict

from qtpy.QtCore import Slot
from qtpy.QtWidgets import (
    QComboBox,
    QLineEdit,
    QPushButton,
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QGridLayout,
)

from MDANSE.Framework.AtomMapping import guess_element
from MDANSE.Chemistry import ATOMS_DATABASE

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase
from MDANSE_GUI.InputWidgets.InputFileWidget import InputFileWidget


class AtomMappingHelperDialog(QDialog):
    """The atom mapping helper dialog used to help generate the atom
    map setting JSON string.
    """

    def __init__(
        self, field_widget: InputFileWidget, field: QLineEdit, parent, *args, **kwargs
    ):
        """
        Parameters
        ----------
        field_widget: InputFileWidget
            The input file widget which is used to obtain the group and
            atom labels.
        field : QLineEdit
            The QLineEdit field that will need to be updated when
            applying the setting.
        """
        super().__init__(parent, *args, **kwargs)
        self.setWindowTitle("Atom mapping helper")
        self.min_width = kwargs.get("min_width", 400)
        self.resize(self.min_width, self.height())
        self.setMinimumWidth(self.min_width)
        self._field = field

        self._file_widget = field_widget
        self._file_widget._field.textChanged.connect(self.update_helper)

        self.layout = QVBoxLayout()
        header_layout = QHBoxLayout()
        header_layout.addWidget(QLabel("Group"))
        header_layout.addWidget(QLabel("Atom Label"))
        header_layout.addWidget(QLabel("Element"))

        self.mapping_widgets = []
        self.mapping_layout = QGridLayout()

        button_layout = QHBoxLayout()
        apply = QPushButton("Apply")
        auto = QPushButton("Auto Fill")
        close = QPushButton("Close")
        apply.clicked.connect(self.apply)
        auto.clicked.connect(self.auto_fill)
        close.clicked.connect(self.close)
        button_layout.addWidget(apply)
        button_layout.addWidget(auto)
        button_layout.addWidget(close)

        self.layout.addLayout(header_layout)
        self.layout.addLayout(self.mapping_layout)
        self.layout.addLayout(button_layout)
        self.setLayout(self.layout)

        self.all_symbols = ATOMS_DATABASE.atoms
        self.labels = None

    def update_helper(self) -> None:
        """Get the atoms labels from the file widget and add mapping
        widgets to the helper.
        """
        self.clear_panel()
        if not self._file_widget._configurator.valid:
            return

        self.labels = self._file_widget._configurator.get_atom_labels()
        for i, label in enumerate(self.labels):
            w0 = QLabel(label.grp_label.replace(";", "\n"))
            w1 = QLabel(label.atm_label)
            w2 = QComboBox()
            w2.addItems(self.all_symbols)
            self.mapping_widgets.append((w0, w1, w2))
            self.mapping_layout.addWidget(w0, i, 0)
            self.mapping_layout.addWidget(w1, i, 1)
            self.mapping_layout.addWidget(w2, i, 2)

        self.auto_fill()

    def clear_panel(self) -> None:
        """Clear the widgets so that it leaves an empty mapping
        layout.
        """
        for ws in self.mapping_widgets:
            for w in ws:
                w.setParent(None)
                self.mapping_layout.removeWidget(w)
        self.mapping_widgets = []

    def auto_fill(self) -> None:
        """Autofill the comboboxes using a simple guess."""
        for i, (_, w1, w2) in enumerate(self.mapping_widgets):
            try:
                label = self.labels[i]
                guess = guess_element(label.atm_label, mass=label.mass)
                idx = self.all_symbols.index(guess)
                w2.setCurrentIndex(idx)
            except AttributeError:
                w2.setCurrentIndex(-1)

    def apply(self) -> None:
        """Convert the selection in the mapping widgets to a JSON string
        and set the field.
        """
        settings = defaultdict(lambda: dict())
        for w0, w1, w2 in self.mapping_widgets:
            settings[w0.text().replace("\n", ";")][w1.text()] = w2.currentText()
        self._field.setText(json.dumps(settings))


class AtomMappingWidget(WidgetBase):
    """The atom mapping widget."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        default_value = "{}"
        self._value = default_value
        self._field = QLineEdit(default_value, self._base)
        self._field.setPlaceholderText(default_value)
        self._field.setMaxLength(2147483647)  # set to the largest possible
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "Replace the chemical elements in the system. The order of symbols is <current>:<new>"
        self._field.setToolTip(tooltip_text)

        # file input widgets should be loaded into the _widgets list
        # before this one
        for widget in self.parent()._widgets:
            if (
                widget._configurator
                is self._configurator._configurable[
                    self._configurator._dependencies["input_file"]
                ]
            ):
                self._file_widget = widget
                break

        self.helper = AtomMappingHelperDialog(
            self._file_widget, self._field, self._base
        )
        self.helper_button = QPushButton("Atom mapping helper", self._base)
        self.helper_button.clicked.connect(self.helper_dialog)
        self.helper_button.setEnabled(False)
        self._field.textChanged.connect(self.updateValue)
        self._default_value = default_value
        self._layout.addWidget(self._field)
        self._layout.addWidget(self.helper_button)
        self._file_widget._field.textChanged.connect(self.update_helper_button)
        self.update_labels()
        self.updateValue()

    @Slot()
    def update_helper_button(self) -> None:
        """Enables the helper button when the file widget has a valid
        file input.
        """
        if self._file_widget._configurator.valid:
            self.helper_button.setEnabled(True)
        else:
            self.helper_button.setEnabled(False)
        self.helper.apply()

    @Slot()
    def helper_dialog(self) -> None:
        """Opens the helper dialog."""
        if self.helper.isVisible():
            self.helper.close()
        else:
            self.helper.show()

    def get_widget_value(self) -> str:
        """
        Returns
        -------
        str
            The JSON string map setting.
        """
        mapping_string = self._field.text()
        if len(mapping_string) < 1:
            self._empty = True
            return self._default_value
        else:
            self._empty = False
        return mapping_string



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import Union, Iterator
from itertools import count, groupby
from qtpy.QtCore import Qt, QEvent, QObject
from qtpy.QtGui import QStandardItem
from qtpy.QtWidgets import QComboBox


class CheckableComboBox(QComboBox):
    """A multi-select checkable combobox"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setEditable(True)
        self.lineEdit().setReadOnly(True)
        self.view().viewport().installEventFilter(self)
        self.view().setAutoScroll(False)
        # it's faster to access the items through this python list than
        # through self.model().item(idx)
        self._items = []
        self.addItem("select all", underline=True)
        self.lineEdit().setText("")

    def eventFilter(self, a0: Union[QObject, None], a1: Union[QEvent, None]) -> bool:
        """Updates the check state of the items and the lineEdit.

        Parameters
        ----------
        a0 : QObject or None
            A QT object.
        a1 : QEvent or None
            A QT event.
        """
        if a0 == self.view().viewport() and a1.type() == QEvent.MouseButtonRelease:
            idx = self.view().indexAt(a1.pos())
            item = self.model().item(idx.row())

            if item.checkState() == Qt.Checked:
                check_uncheck = Qt.Unchecked
            else:
                check_uncheck = Qt.Checked

            if idx.row() == 0:
                # need to block signals temporarily otherwise as we
                # need to make a change on all the items which could
                # cause alot of signals to be emitted
                self.model().blockSignals(True)
                for i in self.getItems():
                    i.setCheckState(check_uncheck)
                self.model().blockSignals(False)
                item.setCheckState(check_uncheck)
            else:
                item.setCheckState(check_uncheck)
                self.update_all_selected()

            self.update_line_edit()
            return True

        return super().eventFilter(a0, a1)

    def update_all_selected(self):
        """check/uncheck select all since everything is/isn't selected."""
        if all([i.checkState() == Qt.Checked for i in self.getItems()]):
            self.model().item(0).setCheckState(Qt.Checked)
        else:
            self.model().item(0).setCheckState(Qt.Unchecked)

    def addItems(self, texts: list[str]) -> None:
        """
        Parameters
        ----------
        texts : list[str]
            A list of items texts to add.
        """
        for text in texts:
            self.addItem(text)

    def configure_using_default(self):
        """This is too complex to have a default value"""

    def addItem(self, text: str, bold=False, italic=False, underline=False) -> None:
        """
        Parameters
        ----------
        text : str
            The text of the item to add.
        """
        item = QStandardItem()
        item.setText(text)
        item.setEnabled(True)
        item.setCheckable(True)
        if bold or italic or underline:
            font = item.font()
            if bold:
                font.setBold(bold)
            if italic:
                font.setItalic(italic)
            if underline:
                font.setUnderline(underline)
            item.setFont(font)
        self.model().appendRow(item)
        self._items.append(item)

    def getItems(self) -> Iterator[QStandardItem]:
        """
        Yields
        ------
        QStandardItem
            Yields the items in the combobox except for the zeroth
            item because that is the select all item.
        """
        for i in range(self.model().rowCount()):
            if i == 0:  # skips the select all item
                continue
            yield self._items[i]

    def check_items_castable_to_int(self) -> bool:
        """
        Returns
        -------
        bool
            Returns true if the text of all items can be cast to int.
        """
        try:
            [int(i.text()) for i in self.getItems()]
            return True
        except ValueError:
            return False

    def update_line_edit(self) -> None:
        """Updates the lineEdit text of the combobox."""
        vals = []
        for item in self.getItems():
            if item.checkState() == Qt.Checked:
                vals.append(item.text())
        if self.check_items_castable_to_int():
            vals = [int(i) for i in vals]
            # changes for example 1,2,3,5,6,7,9,10 -> 1-3,5-7,9-10
            gr = (list(x) for _, x in groupby(vals, lambda x, c=count(): next(c) - x))
            text = ",".join("-".join(map(str, (g[0], g[-1])[: len(g)])) for g in gr)
            self.lineEdit().setText(text)
        else:
            self.lineEdit().setText(",".join(vals))

    def set_default(self, default: str) -> None:
        """Checks the item with the text equal to the default parameter.

        Parameters
        ----------
        default : str
            Parameter used to check items in the combobox.
        """
        model = self.model()
        for row_number in range(model.rowCount()):
            index = model.index(row_number, 0)
            item = model.itemFromIndex(index)
            text = model.data(index)
            if text == default:
                item.setCheckState(Qt.Checked)
            else:
                item.setCheckState(Qt.Unchecked)
        self.update_line_edit()

    def checked_values(self) -> list[str]:
        """
        Returns
        -------
        list[str]
            List of items texts that are checked.
        """
        result = []
        model = self.model()
        for row_number in range(model.rowCount()):
            if row_number == 0:
                continue
            index = model.index(row_number, 0)
            item = model.itemFromIndex(index)
            if item.checkState() == Qt.Checked:
                text = model.data(index)
                result.append(text)
        return result



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import time

from logging.handlers import QueueHandler
from multiprocessing import Queue, Process, Event
from multiprocessing.connection import Connection

from MDANSE.MLogging import LOG
from MDANSE.Framework.Jobs.IJob import IJob

from MDANSE_GUI.Subprocess.JobStatusProcess import JobStatusProcess


class Subprocess(Process):

    def __init__(self, *args, **kwargs):
        super().__init__()
        job_name = kwargs.get("job_name")
        self._job_parameters = kwargs.get("job_parameters")
        sending_pipe = kwargs.get("pipe")
        self.queue_0 = Queue()
        self.queue_1 = Queue()
        pause_event = kwargs.get("pause_event")
        self.log_queue = kwargs.get("log_queue")
        self.construct_job(
            job_name, sending_pipe, self.queue_0, self.queue_1, pause_event
        )

    def construct_job(
        self,
        job: str,
        pipe: Connection,
        queue_0: "Queue",
        queue_1: "Queue",
        pause_event: "Event",
    ):
        job_instance = IJob.create(job)
        job_instance.build_configuration()
        status = JobStatusProcess(pipe, queue_0, queue_1, pause_event)
        job_instance._status = status
        self._job_instance = job_instance

    def run(self):
        queue_handler = QueueHandler(self.log_queue)
        LOG.addHandler(queue_handler)
        LOG.info("Running job")
        self._job_instance.run(self._job_parameters)
        LOG.removeHandler(queue_handler)

    def terminate(self):
        """Send out a terminate message to IJob so that it knows to
        terminate and join its subprocesses. We need to do this
        before this subprocess terminates itself.
        """
        if "running_mode" not in self._job_parameters:
            super().terminate()
            return

        if self._job_parameters["running_mode"][0] == "multicore":
            if self.queue_0.get() != "started":
                raise RuntimeError(
                    "For some reason we received a messaged which wasn't "
                    "'started' this is unexpected."
                )
            self.queue_1.put("terminate")
            # Wait until IJob has received the terminate message.
            while not self.queue_1.empty():
                time.sleep(0.1)
            if self.queue_0.get() != "terminated":
                raise RuntimeError(
                    "For some reason we received a messaged which wasn't "
                    "'terminated' this is unexpected."
                )
        super().terminate()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import glob
import os


def path_to_module(path, stop=""):
    path, _ = os.path.splitext(path)

    splitted_path = path.split(os.sep)

    try:
        idx = splitted_path[::-1].index(stop)
    except ValueError:
        idx = 0
    finally:
        module = ".".join(splitted_path[len(splitted_path) - 1 - idx :])

    return module


# # Import all classes in this directory so that classes with @register_reader are registered.
# pwd = os.path.dirname(__file__)
# for x in glob.glob(os.path.join(pwd, '*.py')):
#     mod, _ = os.path.splitext(x)
#     mod = os.path.basename(mod)
#     if mod.startswith('__'):
#         continue

#     module = path_to_module(x, stop="waterstay")
#     __import__(module, globals(), locals())

# __all__ = ['REGISTERED_READERS']



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import (
    QLineEdit,
    QPushButton,
    QVBoxLayout,
    QHBoxLayout,
    QGroupBox,
    QLabel,
    QTextEdit,
    QWidget,
)
from qtpy.QtGui import QDoubleValidator

from MDANSE.Framework.Configurators.PartialChargeConfigurator import PartialChargeMapper
from MDANSE.Framework.InputData.HDFTrajectoryInputData import HDFTrajectoryInputData
from MDANSE_GUI.InputWidgets.AtomSelectionWidget import AtomSelectionWidget
from MDANSE_GUI.InputWidgets.AtomSelectionWidget import SelectionHelper


class ChargeHelper(SelectionHelper):
    """Generates a string that specifies the partial charge mapping.

    Attributes
    ----------
    _helper_title : str
        The title of the helper dialog window.
    """

    _helper_title = "Partial charge helper"

    def __init__(
        self,
        mapper: PartialChargeMapper,
        traj_data: tuple[str, HDFTrajectoryInputData],
        field: QLineEdit,
        parent,
        *args,
        **kwargs,
    ):
        """
        Parameters
        ----------
        mapper : PartialChargeMapper
            The charge mapper initialized with the current chemical
            system.
        traj_data : tuple[str, HDFTrajectoryInputData]
            A tuple of the trajectory data used to load the 3D viewer.
        field : QLineEdit
            The QLineEdit field that will need to be updated when
            applying the setting.
        """
        self.mapper = mapper
        self.charge_textbox = QTextEdit()
        self.charge_textbox.setReadOnly(True)
        self.charge_qline = QLineEdit()
        self.charge_qline.setValidator(QDoubleValidator())
        self.mapper.selector.settings["all"] = False
        super().__init__(mapper.selector, traj_data, field, parent, *args, **kwargs)
        self.all_selection = False
        self.update_charge_textbox()

    def right_widgets(self) -> list[QWidget]:
        """Add the charge textbox to the right widgets.

        Returns
        -------
        list[QWidget]
            List of QWidgets to add to the right layout from
            create_layouts.
        """
        widgets = super().right_widgets()
        return widgets + [self.charge_textbox]

    def left_widgets(self) -> list[QWidget]:
        """Adds a partial charge widget to the selection helper.

        Returns
        -------
        list[QWidget]
            List of QWidgets to add to the first layout from
            create_layouts.
        """
        widgets = super().left_widgets()
        partial_charge = QGroupBox("partial_charge")
        charge_layout = QVBoxLayout()

        combo_layout = QHBoxLayout()
        label = QLabel("Set charge of selection to:")

        combo_layout.addWidget(label)
        combo_layout.addWidget(self.charge_qline)
        charge_layout.addLayout(combo_layout)

        apply_charge = QPushButton("Apply")
        charge_layout.addWidget(apply_charge)
        apply_charge.clicked.connect(self.apply_charges)

        partial_charge.setLayout(charge_layout)
        return widgets + [partial_charge]

    def apply_charges(self) -> None:
        """With the selection and the charge choice update the charge
        mapper and update the charge textbox with the new partial
        charge map info.
        """
        try:
            charge = float(self.charge_qline.text())
        except ValueError:
            # probably an empty QLineEdit box
            return
        self.mapper.update_charges(self.full_settings, charge)
        self.update_charge_textbox()

    def update_charge_textbox(self) -> None:
        """Update the partial charge textbox with the current charge
        mapper setting information.
        """
        map = self.mapper.get_full_setting()

        text = [f"Partial charge mapping:\n"]
        atoms = self.selector.system.atom_list
        for idx, charge in map.items():
            text.append(f"{idx}  ({atoms[idx].full_name}) -> {charge}\n")

        self.charge_textbox.setText("".join(text))

    def reset(self):
        """Reset the mapper so that no charges are set."""
        super().reset()
        self.mapper.reset_setting()
        self.update_charge_textbox()

    def apply(self) -> None:
        """Set the field of the PartialChargeWidget to the currently
        chosen setting in this widget.
        """
        self._field.setText(self.mapper.get_json_setting())


class PartialChargeWidget(AtomSelectionWidget):
    """The partial charge widget."""

    _push_button_text = "Partial charge helper"
    _default_value = "{}"
    _tooltip_text = "Specify the partial charges that will be used in the analysis. The input is a JSON string, and can be created using the helper dialog."

    def create_helper(
        self, traj_data: tuple[str, HDFTrajectoryInputData]
    ) -> ChargeHelper:
        """
        Parameters
        ----------
        traj_data : tuple[str, HDFTrajectoryInputData]
            A tuple of the trajectory data used to load the 3D viewer.

        Returns
        -------
        ChargeHelper
            Create and return the partial charge helper QDialog.
        """
        mapper = self._configurator.get_charge_mapper()
        return ChargeHelper(mapper, traj_data, self._field, self._base)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#


from qtpy.QtWidgets import (
    QWidget,
    QLineEdit,
    QHBoxLayout,
    QLabel,
)
from qtpy.QtGui import QDoubleValidator, QIntValidator
from qtpy.QtCore import Signal, Slot, Qt, QSortFilterProxyModel


class VectorWidget(QWidget):

    value_changed = Signal(list)

    def __init__(self, *args, dtype=None, **kwargs) -> None:
        label_text = kwargs.pop("label", "Vector input")
        super().__init__(*args, **kwargs)
        if dtype is "int":
            self._dtype = "int"
        else:
            self._dtype = "float"
        layout = QHBoxLayout(self)
        self.setLayout(layout)
        layout.addWidget(QLabel(label_text, self))
        self._fields = []
        for _ in range(3):
            temp = QLineEdit(self)
            if self._dtype == "int":
                temp.setValidator(QIntValidator())
                temp.setText("0")
            elif self._dtype == "float":
                temp.setValidator(QDoubleValidator())
                temp.setText("0.0")
            self._fields.append(temp)
            layout.addWidget(temp)
        self.collect_values()
        for temp in self._fields:
            temp.textChanged.connect(self.collect_values)

    @Slot()
    def collect_values(self):
        result = []
        for temp in self._fields:
            strvalue = temp.text()
            if self._dtype == "int":
                value = int(strvalue)
            elif self._dtype == "float":
                value = float(strvalue)
            result.append(value)
        self._value = result
        self.value_changed.emit(result)

    def setText(self, input_string: str):
        if self._dtype == "int":
            converter = int
        elif self._dtype == "float":
            converter = float
        try:
            self.set_value(
                [converter(x) for x in input_string.strip("[]()").split(",")]
            )
        except:
            self.set_value([0, 0, 0])

    def text(self):
        return ",".join([field.text() for field in self._fields])

    @Slot(object)
    def set_value(self, new_value):
        for n, value in enumerate(new_value):
            self._fields[n].setText(str(value))
